@@ -17,6 +17,11 @@ use build::probe;
 sub defined_or($$) {
     defined $_[0] ? $_[0] : $_[1]
 }
+# This allows us to run on perls without List::Util
+sub uniq {
+    my %s; grep { !$s{$_}++ } @_;
+}
+
 
 my $NAME    = 'moar';
 my $GENLIST = 'build/gen.list';
@@ -129,7 +134,7 @@ if (open(my $fh, '<', 'VERSION')) {
     close($fh);
 }
 # .git is a file and not a directory in submodule
-if (-e '.git' && open(my $GIT, '-|', "git describe")) {
+if (-e '.git' && open(my $GIT, '-|', 'git describe --tags "--match=20*"')) {
     $VERSION = <$GIT>;
     close($GIT);
 }
@@ -160,6 +165,8 @@ $config{ldoptiflags}  = $config{ccoptiflags} unless defined $config{ldoptiflags}
 $config{lddebugflags} = $config{ccdebugflags} unless defined $config{lddebugflags};
 $config{ldinstflags}  = $config{ccinstflags} unless defined $config{ldinstflags};
 
+$config{as}           = $config{cc} unless defined $config{as};
+
 # Probe the compiler.
 build::probe::compiler_usability(\%config, \%defaults);
 
@@ -362,6 +369,8 @@ $config{cc} eq 'clang'
 : $config{cc} eq 'gcc'
     ? '-fopt-info-vec-optimized'
     : die if $args{'show-autovect'};
+push @cflags, $config{ccjitflags} if $args{jit};
+
 if (exists $args{'show-autovect-failed'}) {
     push @cflags, '-Rpass-missed=loop-vectorize' if $config{cc} eq 'clang';
     push @cflags, ("-ftree-vectorizer-verbose=" . ($args{'show-autovect-failed'} || 1), "-fopt-info-vec-missed")
@@ -381,7 +390,7 @@ push @cflags, '-DHAVE_TELEMEH' if $args{telemeh};
 push @cflags, '-DWORDS_BIGENDIAN' if $config{be}; # 3rdparty/sha1 needs it and it isnt set on mips;
 push @cflags, $ENV{CFLAGS} if $ENV{CFLAGS};
 push @cflags, $ENV{CPPFLAGS} if $ENV{CPPFLAGS};
-$config{cflags} = join ' ', @cflags;
+$config{cflags} = join ' ', uniq(@cflags);
 
 # generate LDFLAGS
 my @ldflags = ($config{ldmiscflags});
