@@ -4,7 +4,7 @@
 static const MVMREPROps MVMHash_this_repr;
 
 MVM_STATIC_INLINE MVMString * get_string_key(MVMThreadContext *tc, MVMObject *key) {
-    if (!key || REPR(key)->ID != MVM_REPR_ID_MVMString || !IS_CONCRETE(key))
+    if (MVM_UNLIKELY(!key || REPR(key)->ID != MVM_REPR_ID_MVMString || !IS_CONCRETE(key)))
         MVM_exception_throw_adhoc(tc, "MVMHash representation requires MVMString keys");
     return (MVMString *)key;
 }
@@ -27,29 +27,35 @@ static MVMObject * type_object_for(MVMThreadContext *tc, MVMObject *HOW) {
 static void copy_to(MVMThreadContext *tc, MVMSTable *st, void *src, MVMObject *dest_root, void *dest) {
     MVMHashBody *src_body  = (MVMHashBody *)src;
     MVMHashBody *dest_body = (MVMHashBody *)dest;
-    MVMHashEntry *current  = NULL, *tmp = NULL;
-    unsigned bucket_tmp;
+    MVMHashEntry *current  = NULL;
 
     /* NOTE: if we really wanted to, we could avoid rehashing... */
-    HASH_ITER(hash_handle, src_body->hash_head, current, tmp, bucket_tmp) {
+    HASH_ITER_FAST(tc, hash_handle, src_body->hash_head, current, {
         MVMHashEntry *new_entry = MVM_fixed_size_alloc(tc, tc->instance->fsa,
             sizeof(MVMHashEntry));
         MVMString *key = MVM_HASH_KEY(current);
         MVM_ASSIGN_REF(tc, &(dest_root->header), new_entry->value, current->value);
         MVM_HASH_BIND(tc, dest_body->hash_head, key, new_entry);
         MVM_gc_write_barrier(tc, &(dest_root->header), &(key->common.header));
-    }
+    });
 }
 
 /* Adds held objects to the GC worklist. */
-static void gc_mark(MVMThreadContext *tc, MVMSTable *st, void *data, MVMGCWorklist *worklist) {
+static void MVMHash_gc_mark(MVMThreadContext *tc, MVMSTable *st, void *data, MVMGCWorklist *worklist) {
     MVMHashBody     *body = (MVMHashBody *)data;
-    MVMHashEntry *current = NULL, *tmp = NULL;
-    unsigned bucket_tmp;
-
-    HASH_ITER(hash_handle, body->hash_head, current, tmp, bucket_tmp) {
-        MVM_gc_worklist_add(tc, worklist, &current->hash_handle.key);
-        MVM_gc_worklist_add(tc, worklist, &current->value);
+    MVMHashEntry *current = NULL;
+    MVM_gc_worklist_presize_for(tc, worklist, 2 * HASH_CNT(hash_handle, body->hash_head));
+    if (worklist->include_gen2) {
+        HASH_ITER_FAST(tc, hash_handle, body->hash_head, current, {
+            MVM_gc_worklist_add_include_gen2_nocheck(tc, worklist, &current->hash_handle.key);
+            MVM_gc_worklist_add_include_gen2_nocheck(tc, worklist, &current->value);
+        });
+    }
+    else {
+        HASH_ITER_FAST(tc, hash_handle, body->hash_head, current, {
+            MVM_gc_worklist_add_no_include_gen2_nocheck(tc, worklist, &current->hash_handle.key);
+            MVM_gc_worklist_add_no_include_gen2_nocheck(tc, worklist, &current->value);
+        });
     }
 }
 
@@ -57,13 +63,12 @@ static void gc_mark(MVMThreadContext *tc, MVMSTable *st, void *data, MVMGCWorkli
 static void gc_free(MVMThreadContext *tc, MVMObject *obj) {
     MVMHash *h = (MVMHash *)obj;
     MVMHashEntry *current = NULL, *tmp = NULL;
-    unsigned bucket_tmp;
-    HASH_ITER(hash_handle, h->body.hash_head, current, tmp, bucket_tmp) {
+    HASH_ITER_FAST(tc, hash_handle, h->body.hash_head, current, {
         if (current != h->body.hash_head)
             MVM_fixed_size_free(tc, tc->instance->fsa, sizeof(MVMHashEntry), current);
-    }
+    });
     tmp = h->body.hash_head;
-    HASH_CLEAR(hash_handle, h->body.hash_head);
+    HASH_CLEAR(tc, hash_handle, h->body.hash_head);
     if (tmp)
         MVM_fixed_size_free(tc, tc->instance->fsa, sizeof(MVMHashEntry), tmp);
 }
@@ -72,19 +77,22 @@ static void at_key(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *d
     MVMHashBody   *body = (MVMHashBody *)data;
     MVMHashEntry *entry = NULL;
     MVM_HASH_GET(tc, body->hash_head, get_string_key(tc, key_obj), entry);
-    if (kind == MVM_reg_obj)
+    if (MVM_LIKELY(kind == MVM_reg_obj))
         result->o = entry != NULL ? entry->value : tc->instance->VMNull;
     else
         MVM_exception_throw_adhoc(tc,
             "MVMHash representation does not support native type storage");
 }
+void MVMHash_at_key(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *data, MVMObject *key_obj, MVMRegister *result, MVMuint16 kind) {
+    at_key(tc, st, root, data, key_obj, result, kind);
+}
 
 static void bind_key(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *data, MVMObject *key_obj, MVMRegister value, MVMuint16 kind) {
     MVMHashBody   *body = (MVMHashBody *)data;
     MVMHashEntry *entry = NULL;
 
     MVMString *key = get_string_key(tc, key_obj);
-    if (kind != MVM_reg_obj)
+    if (MVM_UNLIKELY(kind != MVM_reg_obj))
         MVM_exception_throw_adhoc(tc,
             "MVMHash representation does not support native type storage");
 
@@ -101,7 +109,9 @@ static void bind_key(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void
         MVM_ASSIGN_REF(tc, &(root->header), entry->value, value.o);
     }
 }
-
+void MVMHash_bind_key(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *data, MVMObject *key_obj, MVMRegister value, MVMuint16 kind) {
+    bind_key(tc, st, root, data, key_obj, value, kind);
+}
 static MVMuint64 elems(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *data) {
     MVMHashBody *body = (MVMHashBody *)data;
     return HASH_CNT(hash_handle, body->hash_head);
@@ -117,10 +127,9 @@ static MVMint64 exists_key(MVMThreadContext *tc, MVMSTable *st, MVMObject *root,
 static void delete_key(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *data, MVMObject *key_obj) {
     MVMHashBody *body = (MVMHashBody *)data;
     MVMString *key = get_string_key(tc, key_obj);
-    MVMHashEntry *old_entry = NULL;
-    MVM_HASH_GET(tc, body->hash_head, key, old_entry);
+    MVMHashEntry *old_entry = NULL, *prev_entry = NULL;
+    HASH_FIND_VM_STR_AND_DELETE(tc, hash_handle, body->hash_head, key, old_entry, prev_entry);
     if (old_entry) {
-        HASH_DELETE(hash_handle, body->hash_head, old_entry);
         MVM_fixed_size_free(tc, tc->instance->fsa,
             sizeof(MVMHashEntry), old_entry);
     }
@@ -178,15 +187,14 @@ static int cmp_strings(const void *s1, const void *s2) {
 }
 static void serialize(MVMThreadContext *tc, MVMSTable *st, void *data, MVMSerializationWriter *writer) {
     MVMHashBody *body = (MVMHashBody *)data;
-    MVMHashEntry *current = NULL, *tmp = NULL;
+    MVMHashEntry *current = NULL;
     MVMuint64 elems = HASH_CNT(hash_handle, body->hash_head);
     MVMString **keys = MVM_malloc(sizeof(MVMString *) * elems);
-    unsigned bucket_tmp;
     MVMuint64 i = 0;
     MVM_serialization_write_int(tc, writer, elems);
-    HASH_ITER(hash_handle, body->hash_head, current, tmp, bucket_tmp) {
+    HASH_ITER_FAST(tc, hash_handle, body->hash_head, current, {
         keys[i++] = MVM_HASH_KEY(current);
-    }
+    });
     cmp_tc = tc;
     qsort(keys, elems, sizeof(MVMString*), cmp_strings);
     for (i = 0; i < elems; i++) {
@@ -215,7 +223,7 @@ static void spesh(MVMThreadContext *tc, MVMSTable *st, MVMSpeshGraph *g, MVMSpes
             ins->operands[0]         = target;
             ins->operands[1].lit_i16 = sizeof(MVMHash);
             ins->operands[2].lit_i16 = MVM_spesh_add_spesh_slot(tc, g, (MVMCollectable *)st);
-            MVM_spesh_get_facts(tc, g, type)->usages--;
+            MVM_spesh_usages_delete_by_reg(tc, g, type, ins);
         }
         break;
     }
@@ -256,7 +264,7 @@ static const MVMREPROps MVMHash_this_repr = {
     NULL, /* serialize_repr_data */
     NULL, /* deserialize_repr_data */
     deserialize_stable_size,
-    gc_mark,
+    MVMHash_gc_mark,
     gc_free,
     NULL, /* gc_cleanup */
     NULL, /* gc_mark_repr_data */
