@@ -1,22 +1,16 @@
 #include "moar.h"
 
 /* This representation's function pointer table. */
-static const MVMREPROps this_repr;
-
-/* Invocation protocol handler. */
-static void invoke_handler(MVMThreadContext *tc, MVMObject *invokee, MVMCallsite *callsite, MVMRegister *args) {
-    MVM_exception_throw_adhoc(tc, "Cannot invoke static frame object");
-}
+static const MVMREPROps MVMStaticFrame_this_repr;
 
 /* Creates a new type object of this representation, and associates it with
  * the given HOW. Also sets the invocation protocol handler in the STable. */
 static MVMObject * type_object_for(MVMThreadContext *tc, MVMObject *HOW) {
-    MVMSTable *st = MVM_gc_allocate_stable(tc, &this_repr, HOW);
+    MVMSTable *st = MVM_gc_allocate_stable(tc, &MVMStaticFrame_this_repr, HOW);
 
     MVMROOT(tc, st, {
         MVMObject *obj = MVM_gc_allocate_type_object(tc, st);
         MVM_ASSIGN_REF(tc, &(st->header), st->WHAT, obj);
-        st->invoke = invoke_handler;
         st->size = sizeof(MVMStaticFrame);
     });
 
@@ -59,27 +53,23 @@ static void copy_to(MVMThreadContext *tc, MVMSTable *st, void *src, MVMObject *d
         MVMuint16 *local_types = MVM_malloc(sizeof(MVMuint16) * src_body->num_locals);
         MVMuint16 *lexical_types = MVM_malloc(sizeof(MVMuint16) * src_body->num_lexicals);
         memcpy(local_types, src_body->local_types, sizeof(MVMuint16) * src_body->num_locals);
-        memcpy(lexical_types, src_body->lexical_types, sizeof(MVMuint16) * src_body->num_lexicals);
+        if (src_body->num_lexicals)
+            memcpy(lexical_types, src_body->lexical_types,
+                sizeof(MVMuint16) * src_body->num_lexicals);
         dest_body->local_types = local_types;
         dest_body->lexical_types = lexical_types;
     }
     {
         MVMLexicalRegistry *current, *tmp;
-        unsigned bucket_tmp;
 
         /* NOTE: if we really wanted to, we could avoid rehashing... */
-        HASH_ITER(hash_handle, src_body->lexical_names, current, tmp, bucket_tmp) {
-            size_t klen;
-            void *kdata;
+        HASH_ITER_FAST(tc, hash_handle, src_body->lexical_names, current, {
             MVMLexicalRegistry *new_entry = MVM_malloc(sizeof(MVMLexicalRegistry));
-
             /* don't need to clone the string */
             MVM_ASSIGN_REF(tc, &(dest_root->header), new_entry->key, current->key);
             new_entry->value = current->value;
-
-            MVM_HASH_EXTRACT_KEY(tc, &kdata, &klen, current->key, "really broken")
-            HASH_ADD_KEYPTR(hash_handle, dest_body->lexical_names, kdata, klen, new_entry);
-        }
+            MVM_HASH_BIND(tc, dest_body->lexical_names, current->key, new_entry);
+        });
     }
 
     /* Static environment needs to be copied, and any objects WB'd. */
@@ -110,9 +100,10 @@ static void copy_to(MVMThreadContext *tc, MVMSTable *st, void *src, MVMObject *d
 
     dest_body->num_handlers = src_body->num_handlers;
     dest_body->handlers     = MVM_malloc(src_body->num_handlers * sizeof(MVMFrameHandler));
-    memcpy(dest_body->handlers, src_body->handlers, src_body->num_handlers * sizeof(MVMFrameHandler));
+    if (src_body->num_handlers)
+        memcpy(dest_body->handlers, src_body->handlers,
+            src_body->num_handlers * sizeof(MVMFrameHandler));
     dest_body->instrumentation_level = 0;
-    dest_body->pool_index            = src_body->pool_index;
     dest_body->num_annotations       = src_body->num_annotations;
     dest_body->annotations_data      = src_body->annotations_data;
     dest_body->fully_deserialized    = 1;
@@ -121,8 +112,7 @@ static void copy_to(MVMThreadContext *tc, MVMSTable *st, void *src, MVMObject *d
 /* Adds held objects to the GC worklist. */
 static void gc_mark(MVMThreadContext *tc, MVMSTable *st, void *data, MVMGCWorklist *worklist) {
     MVMStaticFrameBody *body = (MVMStaticFrameBody *)data;
-    MVMLexicalRegistry *current, *tmp;
-    unsigned bucket_tmp;
+    MVMLexicalRegistry *current;
 
     /* mvmobjects */
     MVM_gc_worklist_add(tc, worklist, &body->cu);
@@ -136,9 +126,10 @@ static void gc_mark(MVMThreadContext *tc, MVMSTable *st, void *data, MVMGCWorkli
         return;
 
     /* lexical names hash keys */
-    HASH_ITER(hash_handle, body->lexical_names, current, tmp, bucket_tmp) {
+    HASH_ITER_FAST(tc, hash_handle, body->lexical_names, current, {
+        MVM_gc_worklist_add(tc, worklist, &current->hash_handle.key);
         MVM_gc_worklist_add(tc, worklist, &current->key);
-    }
+    });
 
     /* static env */
     if (body->static_env) {
@@ -150,30 +141,14 @@ static void gc_mark(MVMThreadContext *tc, MVMSTable *st, void *data, MVMGCWorkli
                 MVM_gc_worklist_add(tc, worklist, &body->static_env[i].o);
     }
 
-    /* Spesh slots. */
-    if (body->num_spesh_candidates) {
-        MVMint32 i, j;
-        for (i = 0; i < body->num_spesh_candidates; i++) {
-            for (j = 0; j < body->spesh_candidates[i].num_guards; j++)
-                MVM_gc_worklist_add(tc, worklist, &body->spesh_candidates[i].guards[j].match);
-            for (j = 0; j < body->spesh_candidates[i].num_spesh_slots; j++)
-                MVM_gc_worklist_add(tc, worklist, &body->spesh_candidates[i].spesh_slots[j]);
-            if (body->spesh_candidates[i].log_slots)
-                for (j = 0; j < body->spesh_candidates[i].num_log_slots * MVM_SPESH_LOG_RUNS; j++)
-                    MVM_gc_worklist_add(tc, worklist, &body->spesh_candidates[i].log_slots[j]);
-            for (j = 0; j < body->spesh_candidates[i].num_inlines; j++)
-                MVM_gc_worklist_add(tc, worklist, &body->spesh_candidates[i].inlines[j].code);
-            if (body->spesh_candidates[i].sg)
-                MVM_spesh_graph_mark(tc, body->spesh_candidates[i].sg, worklist);
-        }
-    }
+    /* Spesh. */
+    MVM_gc_worklist_add(tc, worklist, &body->spesh);
 }
 
 /* Called by the VM in order to free memory associated with this object. */
 static void gc_free(MVMThreadContext *tc, MVMObject *obj) {
     MVMStaticFrame *sf = (MVMStaticFrame *)obj;
     MVMStaticFrameBody *body = &sf->body;
-    MVMint32 i;
     if (body->orig_bytecode != body->bytecode) {
         MVM_free(body->bytecode);
         body->bytecode = body->orig_bytecode;
@@ -184,16 +159,13 @@ static void gc_free(MVMThreadContext *tc, MVMObject *obj) {
     if (!body->fully_deserialized)
         return;
     MVM_free(body->handlers);
+    MVM_free(body->work_initial);
     MVM_free(body->static_env);
     MVM_free(body->static_env_flags);
     MVM_free(body->local_types);
     MVM_free(body->lexical_types);
     MVM_free(body->lexical_names_list);
-    MVM_HASH_DESTROY(hash_handle, MVMLexicalRegistry, body->lexical_names);
-
-    for (i = 0; i < body->num_spesh_candidates; i++)
-        MVM_spesh_candidate_destroy(tc, &body->spesh_candidates[i]);
-    MVM_free(body->spesh_candidates);
+    MVM_HASH_DESTROY(tc, hash_handle, MVMLexicalRegistry, body->lexical_names);
 }
 
 static const MVMStorageSpec storage_spec = {
@@ -223,8 +195,6 @@ static MVMuint64 unmanaged_size(MVMThreadContext *tc, MVMSTable *st, void *data)
     MVMuint64 size = 0;
 
     if (body->fully_deserialized) {
-        MVMuint32 spesh_idx;
-
         size += sizeof(MVMuint16) * body->num_locals;
         size += sizeof(MVMuint16) * body->num_lexicals;
 
@@ -245,47 +215,6 @@ static MVMuint64 unmanaged_size(MVMThreadContext *tc, MVMSTable *st, void *data)
         size += body->env_size; /* static_env */
         size += body->num_lexicals; /* static_env_flags */
 
-        for (spesh_idx = 0; spesh_idx < body->num_spesh_candidates; spesh_idx++) {
-            MVMSpeshCandidate *cand = &body->spesh_candidates[spesh_idx];
-            size += sizeof(MVMSpeshGuard) * cand->num_guards;
-
-            size += cand->bytecode_size;
-
-            size += sizeof(MVMFrameHandler) * cand->num_handlers;
-
-            size += sizeof(MVMCollectable *) * cand->num_spesh_slots;
-
-            size += sizeof(MVMint32) * cand->num_deopts;
-
-            if (cand->sg)
-                /* XXX we ought to descend into the speshgraph, too. */
-                size += sizeof(MVMSpeshGraph);
-
-            size += sizeof(MVMCollectable *) * cand->num_log_slots;
-
-            size += sizeof(MVMSpeshInline) * cand->num_inlines;
-
-            size += sizeof(MVMuint16) * (cand->num_locals + cand->num_lexicals);
-
-            /* XXX probably don't need to measure the bytecode size here,
-             * as it's probably just a pointer to the same bytecode we have in
-             * the static frame anyway. */
-
-            /* Dive into the jit code */
-            if (cand->jitcode) {
-                MVMJitCode *code = cand->jitcode;
-
-                size += sizeof(MVMJitCode);
-
-                size += sizeof(void *) * code->num_labels;
-
-                size += sizeof(MVMint32) * code->num_bbs;
-                size += sizeof(MVMJitDeopt) * code->num_deopts;
-                size += sizeof(MVMJitInline) * code->num_inlines;
-                size += sizeof(MVMJitHandler) * code->num_handlers;
-            }
-        }
-
         if (body->instrumentation) {
             size += body->instrumentation->uninstrumented_bytecode_size;
             size += body->instrumentation->instrumented_bytecode_size;
@@ -301,8 +230,7 @@ static MVMuint64 unmanaged_size(MVMThreadContext *tc, MVMSTable *st, void *data)
 
 static void describe_refs(MVMThreadContext *tc, MVMHeapSnapshotState *ss, MVMSTable *st, void *data) {
     MVMStaticFrameBody *body = (MVMStaticFrameBody *)data;
-    MVMLexicalRegistry *current, *tmp;
-    unsigned bucket_tmp;
+    MVMLexicalRegistry *current;
 
     MVM_profile_heap_add_collectable_rel_const_cstr(tc, ss,
         (MVMCollectable *)body->cu, "Compilation Unit");
@@ -320,10 +248,10 @@ static void describe_refs(MVMThreadContext *tc, MVMHeapSnapshotState *ss, MVMSTa
         return;
 
     /* lexical names hash keys */
-    HASH_ITER(hash_handle, body->lexical_names, current, tmp, bucket_tmp) {
+    HASH_ITER_FAST(tc, hash_handle, body->lexical_names, current, {
         MVM_profile_heap_add_collectable_rel_const_cstr(tc, ss,
             (MVMCollectable *)current->key, "Lexical name");
-    }
+    });
 
     /* static env */
     if (body->static_env) {
@@ -336,46 +264,17 @@ static void describe_refs(MVMThreadContext *tc, MVMHeapSnapshotState *ss, MVMSTa
                     (MVMCollectable *)body->static_env[i].o, "Static Environment Entry");
     }
 
-    /* Spesh slots. */
-    if (body->num_spesh_candidates) {
-        MVMint32 i, j;
-        for (i = 0; i < body->num_spesh_candidates; i++) {
-            for (j = 0; j < body->spesh_candidates[i].num_guards; j++)
-                MVM_profile_heap_add_collectable_rel_const_cstr(tc, ss,
-                    (MVMCollectable *)body->spesh_candidates[i].guards[j].match,
-                    "Spesh guard match");
-            for (j = 0; j < body->spesh_candidates[i].num_spesh_slots; j++)
-                MVM_profile_heap_add_collectable_rel_const_cstr(tc, ss,
-                    (MVMCollectable *)body->spesh_candidates[i].spesh_slots[j],
-                    "Spesh slot entry");
-            if (body->spesh_candidates[i].log_slots)
-                for (j = 0; j < body->spesh_candidates[i].num_log_slots * MVM_SPESH_LOG_RUNS; j++)
-                MVM_profile_heap_add_collectable_rel_const_cstr(tc, ss,
-                    (MVMCollectable *)body->spesh_candidates[i].log_slots[j],
-                    "Spesh log slots");
-            for (j = 0; j < body->spesh_candidates[i].num_inlines; j++)
-                MVM_profile_heap_add_collectable_rel_const_cstr(tc, ss,
-                    (MVMCollectable *)body->spesh_candidates[i].inlines[j].code,
-                    "Spesh inlined code object");
-            if (body->spesh_candidates[i].sg) {
-                MVMCollectable **c_ptr;
-                MVM_spesh_graph_mark(tc, body->spesh_candidates[i].sg, ss->gcwl);
-                while (c_ptr = MVM_gc_worklist_get(tc, ss->gcwl)) {
-                    MVMCollectable *c = *c_ptr;
-                    MVM_profile_heap_add_collectable_rel_const_cstr(tc, ss, c,
-                        "Object held by spesh graph");
-                }
-            }
-        }
-    }
+    /* Spesh data */
+    MVM_profile_heap_add_collectable_rel_const_cstr(tc, ss,
+        (MVMCollectable *)body->spesh, "Specializer Data");
 }
 
 /* Initializes the representation. */
 const MVMREPROps * MVMStaticFrame_initialize(MVMThreadContext *tc) {
-    return &this_repr;
+    return &MVMStaticFrame_this_repr;
 }
 
-static const MVMREPROps this_repr = {
+static const MVMREPROps MVMStaticFrame_this_repr = {
     type_object_for,
     MVM_gc_allocate_object,
     NULL, /* initialize */
@@ -401,7 +300,6 @@ static const MVMREPROps this_repr = {
     NULL, /* spesh */
     "MVMStaticFrame", /* name */
     MVM_REPR_ID_MVMStaticFrame,
-    0, /* refs_frames */
     unmanaged_size, /* unmanaged_size */
     describe_refs,
 };
