@@ -203,15 +203,15 @@ static void gc_free_repr_data(MVMThreadContext *tc, MVMSTable *st) {
 }
 
 /* Helper for complaining about attribute access errors. */
-MVM_NO_RETURN
-static void no_such_attribute(MVMThreadContext *tc, const char *action, MVMObject *class_handle, MVMString *name, MVMSTable *target_type) {
+MVM_NO_RETURN static void no_such_attribute(MVMThreadContext *tc, const char *action, MVMObject *class_handle, MVMString *name, MVMSTable *target_type) MVM_NO_RETURN_ATTRIBUTE;
+MVM_NO_RETURN static void no_such_attribute(MVMThreadContext *tc, const char *action, MVMObject *class_handle, MVMString *name, MVMSTable *target_type) {
     char *c_name = MVM_string_utf8_encode_C_string(tc, name);
     char *waste[] = { c_name, NULL };
     MVM_exception_throw_adhoc_free(tc, waste, "P6opaque: no such attribute '%s' on type %s in a %s when trying to %s", c_name, MVM_6model_get_debug_name(tc, class_handle), MVM_6model_get_stable_debug_name(tc, target_type), action);
 }
 
-MVM_NO_RETURN
-static void invalid_access_kind(MVMThreadContext *tc, const char *action, MVMObject *class_handle, MVMString *name, const char *kind_desc) {
+MVM_NO_RETURN static void invalid_access_kind(MVMThreadContext *tc, const char *action, MVMObject *class_handle, MVMString *name, const char *kind_desc) MVM_NO_RETURN_ATTRIBUTE;
+MVM_NO_RETURN static void invalid_access_kind(MVMThreadContext *tc, const char *action, MVMObject *class_handle, MVMString *name, const char *kind_desc) {
     char *c_name = MVM_string_utf8_encode_C_string(tc, name);
     char *waste[] = { c_name, NULL };
     MVM_exception_throw_adhoc_free(tc, waste, "P6opaque: invalid %s attribute '%s' in type %s for kind %s", action, c_name, MVM_6model_get_debug_name(tc, class_handle), kind_desc);
@@ -431,7 +431,8 @@ static MVMint64 hint_for(MVMThreadContext *tc, MVMSTable *st, MVMObject *class_k
 /* Gets an architecture atomic sized native integer attribute as an atomic
  * reference. */
 static AO_t * attribute_as_atomic(MVMThreadContext *tc, MVMSTable *st, void *data,
-                                  MVMObject *class_handle, MVMString *name) {
+                                  MVMObject *class_handle, MVMString *name,
+                                  MVMuint16 kind) {
     MVMP6opaqueREPRData *repr_data = (MVMP6opaqueREPRData *)st->REPR_data;
     MVMint64 slot;
     data = MVM_p6opaque_real_data(tc, data);
@@ -440,15 +441,24 @@ static AO_t * attribute_as_atomic(MVMThreadContext *tc, MVMSTable *st, void *dat
             "P6opaque: must compose %s before using get_attribute", MVM_6model_get_stable_debug_name(tc, st));
     slot = try_get_slot(tc, repr_data, class_handle, name);
     if (slot >= 0) {
-        MVMSTable *attr_st = repr_data->flattened_stables[slot];
-        if (attr_st) {
-            const MVMStorageSpec *ss = attr_st->REPR->get_storage_spec(tc, attr_st);
-            if (ss->inlineable && ss->boxed_primitive == MVM_STORAGE_SPEC_BP_INT &&
-                    ss->bits / 8 == sizeof(AO_t))
-                return (AO_t *)((char *)data + repr_data->attribute_offsets[slot]);
+        if (kind == MVM_reg_int64) {
+            MVMSTable *attr_st = repr_data->flattened_stables[slot];
+            if (attr_st) {
+                const MVMStorageSpec *ss = attr_st->REPR->get_storage_spec(tc, attr_st);
+                if (ss->inlineable && ss->boxed_primitive == MVM_STORAGE_SPEC_BP_INT &&
+                        ss->bits / 8 == sizeof(AO_t))
+                    return (AO_t *)((char *)data + repr_data->attribute_offsets[slot]);
+            }
+            MVM_exception_throw_adhoc(tc,
+                "Can only do an atomic integer operation on an atomicint attribute");
+        }
+        else if (kind == MVM_reg_obj) {
+            return (AO_t *)((char *)data + repr_data->attribute_offsets[slot]);
+        }
+        else {
+            MVM_exception_throw_adhoc(tc,
+                "Can only perform atomic operations on object or atomicint attributes");
         }
-        MVM_exception_throw_adhoc(tc,
-            "Can only do an atomic integer operation on an atomicint attribute");
     }
     else {
         no_such_attribute(tc, "get atomic reference to", class_handle, name, st);
@@ -1235,6 +1245,9 @@ static void at_pos(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *d
     del = get_obj_at_offset(data, repr_data->attribute_offsets[repr_data->pos_del_slot]);
     REPR(del)->pos_funcs.at_pos(tc, STABLE(del), del, OBJECT_BODY(del), index, value, kind);
 }
+void MVM_P6opaque_at_pos(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *data, MVMint64 index, MVMRegister *value, MVMuint16 kind) {
+    at_pos(tc, st, root, data, index, value, kind);
+}
 
 static void bind_pos(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *data, MVMint64 index, MVMRegister value, MVMuint16 kind) {
     MVMP6opaqueREPRData *repr_data = (MVMP6opaqueREPRData *)st->REPR_data;
@@ -1405,7 +1418,7 @@ static void spesh(MVMThreadContext *tc, MVMSTable *st, MVMSpeshGraph *g, MVMSpes
             ins->operands[0]         = target;
             ins->operands[1].lit_i16 = st->size;
             ins->operands[2].lit_i16 = MVM_spesh_add_spesh_slot(tc, g, (MVMCollectable *)st);
-            MVM_spesh_get_facts(tc, g, type)->usages--;
+            MVM_spesh_usages_delete_by_reg(tc, g, type, ins);
         }
         break;
     }
@@ -1427,16 +1440,16 @@ static void spesh(MVMThreadContext *tc, MVMSTable *st, MVMSpeshGraph *g, MVMSpes
                         ins->info = MVM_op_get_op(MVM_OP_sp_p6ogetvt_o);
                     }
                     if (opcode == MVM_OP_getattrs_o)
-                        MVM_spesh_get_facts(tc, g, ins->operands[3])->usages--;
-                    MVM_spesh_get_facts(tc, g, ins->operands[2])->usages--;
+                        MVM_spesh_usages_delete_by_reg(tc, g, ins->operands[3], ins);
+                    MVM_spesh_usages_delete_by_reg(tc, g, ins->operands[2], ins);
                     ins->operands[2].lit_i16 = repr_data->attribute_offsets[slot];
                     ins->operands[3].lit_i16 = MVM_spesh_add_spesh_slot(tc, g,
                         (MVMCollectable *)av_value);
                 }
                 else {
                     if (opcode == MVM_OP_getattrs_o)
-                        MVM_spesh_get_facts(tc, g, ins->operands[3])->usages--;
-                    MVM_spesh_get_facts(tc, g, ins->operands[2])->usages--;
+                        MVM_spesh_usages_delete_by_reg(tc, g, ins->operands[3], ins);
+                    MVM_spesh_usages_delete_by_reg(tc, g, ins->operands[2], ins);
                     ins->info = MVM_op_get_op(MVM_OP_sp_p6oget_o);
                     ins->operands[2].lit_i16 = repr_data->attribute_offsets[slot];
                     MVM_spesh_manipulate_remove_handler_successors(tc, bb);
@@ -1458,8 +1471,8 @@ static void spesh(MVMThreadContext *tc, MVMSTable *st, MVMSpeshGraph *g, MVMSpes
                 const MVMStorageSpec *flat_ss = flat_st->REPR->get_storage_spec(tc, flat_st);
                 if (flat_st->REPR->ID == MVM_REPR_ID_P6int && flat_ss->bits == 64) {
                     if (opcode == MVM_OP_getattrs_i)
-                        MVM_spesh_get_facts(tc, g, ins->operands[3])->usages--;
-                    MVM_spesh_get_facts(tc, g, ins->operands[2])->usages--;
+                        MVM_spesh_usages_delete_by_reg(tc, g, ins->operands[3], ins);
+                    MVM_spesh_usages_delete_by_reg(tc, g, ins->operands[2], ins);
                     ins->info = MVM_op_get_op(MVM_OP_sp_p6oget_i);
                     ins->operands[2].lit_i16 = repr_data->attribute_offsets[slot];
                 }
@@ -1480,8 +1493,8 @@ static void spesh(MVMThreadContext *tc, MVMSTable *st, MVMSpeshGraph *g, MVMSpes
                 const MVMStorageSpec *flat_ss = flat_st->REPR->get_storage_spec(tc, flat_st);
                 if (flat_st->REPR->ID == MVM_REPR_ID_P6num && flat_ss->bits == 64) {
                     if (opcode == MVM_OP_getattrs_n)
-                        MVM_spesh_get_facts(tc, g, ins->operands[3])->usages--;
-                    MVM_spesh_get_facts(tc, g, ins->operands[2])->usages--;
+                        MVM_spesh_usages_delete_by_reg(tc, g, ins->operands[3], ins);
+                    MVM_spesh_usages_delete_by_reg(tc, g, ins->operands[2], ins);
                     ins->info = MVM_op_get_op(MVM_OP_sp_p6oget_n);
                     ins->operands[2].lit_i16 = repr_data->attribute_offsets[slot];
                 }
@@ -1501,8 +1514,8 @@ static void spesh(MVMThreadContext *tc, MVMSTable *st, MVMSpeshGraph *g, MVMSpes
                 MVMSTable      *flat_st = repr_data->flattened_stables[slot];
                 if (flat_st->REPR->ID == MVM_REPR_ID_P6str) {
                     if (opcode == MVM_OP_getattrs_s)
-                        MVM_spesh_get_facts(tc, g, ins->operands[3])->usages--;
-                    MVM_spesh_get_facts(tc, g, ins->operands[2])->usages--;
+                        MVM_spesh_usages_delete_by_reg(tc, g, ins->operands[3], ins);
+                    MVM_spesh_usages_delete_by_reg(tc, g, ins->operands[2], ins);
                     ins->info = MVM_op_get_op(MVM_OP_sp_p6oget_s);
                     ins->operands[2].lit_i16 = repr_data->attribute_offsets[slot];
                 }
@@ -1520,8 +1533,8 @@ static void spesh(MVMThreadContext *tc, MVMSTable *st, MVMSpeshGraph *g, MVMSpes
             MVMint64 slot = try_get_slot(tc, repr_data, ch_facts->type, name);
             if (slot >= 0 && !repr_data->flattened_stables[slot]) {
                 if (opcode == MVM_OP_bindattrs_o)
-                    MVM_spesh_get_facts(tc, g, ins->operands[2])->usages--;
-                MVM_spesh_get_facts(tc, g, ins->operands[1])->usages--;
+                    MVM_spesh_usages_delete_by_reg(tc, g, ins->operands[2], ins);
+                MVM_spesh_usages_delete_by_reg(tc, g, ins->operands[1], ins);
                 ins->info = MVM_op_get_op(MVM_OP_sp_p6obind_o);
                 ins->operands[1].lit_i16 = repr_data->attribute_offsets[slot];
                 ins->operands[2] = ins->operands[3];
@@ -1542,8 +1555,8 @@ static void spesh(MVMThreadContext *tc, MVMSTable *st, MVMSpeshGraph *g, MVMSpes
                 const MVMStorageSpec *flat_ss = flat_st->REPR->get_storage_spec(tc, flat_st);
                 if (flat_st->REPR->ID == MVM_REPR_ID_P6int && flat_ss->bits == 64) {
                     if (opcode == MVM_OP_bindattrs_i)
-                        MVM_spesh_get_facts(tc, g, ins->operands[2])->usages--;
-                    MVM_spesh_get_facts(tc, g, ins->operands[1])->usages--;
+                        MVM_spesh_usages_delete_by_reg(tc, g, ins->operands[2], ins);
+                    MVM_spesh_usages_delete_by_reg(tc, g, ins->operands[1], ins);
                     ins->info = MVM_op_get_op(MVM_OP_sp_p6obind_i);
                     ins->operands[1].lit_i16 = repr_data->attribute_offsets[slot];
                     ins->operands[2] = ins->operands[3];
@@ -1565,8 +1578,8 @@ static void spesh(MVMThreadContext *tc, MVMSTable *st, MVMSpeshGraph *g, MVMSpes
                 const MVMStorageSpec *flat_ss = flat_st->REPR->get_storage_spec(tc, flat_st);
                 if (flat_st->REPR->ID == MVM_REPR_ID_P6num && flat_ss->bits == 64) {
                     if (opcode == MVM_OP_bindattrs_n)
-                        MVM_spesh_get_facts(tc, g, ins->operands[2])->usages--;
-                    MVM_spesh_get_facts(tc, g, ins->operands[1])->usages--;
+                        MVM_spesh_usages_delete_by_reg(tc, g, ins->operands[2], ins);
+                    MVM_spesh_usages_delete_by_reg(tc, g, ins->operands[1], ins);
                     ins->info = MVM_op_get_op(MVM_OP_sp_p6obind_n);
                     ins->operands[1].lit_i16 = repr_data->attribute_offsets[slot];
                     ins->operands[2] = ins->operands[3];
@@ -1587,8 +1600,8 @@ static void spesh(MVMThreadContext *tc, MVMSTable *st, MVMSpeshGraph *g, MVMSpes
                 MVMSTable      *flat_st = repr_data->flattened_stables[slot];
                 if (flat_st->REPR->ID == MVM_REPR_ID_P6str) {
                     if (opcode == MVM_OP_bindattrs_s)
-                        MVM_spesh_get_facts(tc, g, ins->operands[2])->usages--;
-                    MVM_spesh_get_facts(tc, g, ins->operands[1])->usages--;
+                        MVM_spesh_usages_delete_by_reg(tc, g, ins->operands[2], ins);
+                    MVM_spesh_usages_delete_by_reg(tc, g, ins->operands[1], ins);
                     ins->info = MVM_op_get_op(MVM_OP_sp_p6obind_s);
                     ins->operands[1].lit_i16 = repr_data->attribute_offsets[slot];
                     ins->operands[2] = ins->operands[3];
