@@ -62,7 +62,7 @@ static void copy_to(MVMThreadContext *tc, MVMSTable *st, void *src, MVMObject *d
 }
 
 /* Adds held objects to the GC worklist. */
-static void gc_mark(MVMThreadContext *tc, MVMSTable *st, void *data, MVMGCWorklist *worklist) {
+static void VMArray_gc_mark(MVMThreadContext *tc, MVMSTable *st, void *data, MVMGCWorklist *worklist) {
     MVMArrayREPRData *repr_data = (MVMArrayREPRData *)st->REPR_data;
     MVMArrayBody     *body      = (MVMArrayBody *)data;
     MVMuint64         elems     = body->elems;
@@ -72,18 +72,28 @@ static void gc_mark(MVMThreadContext *tc, MVMSTable *st, void *data, MVMGCWorkli
         case MVM_ARRAY_OBJ: {
             MVMObject **slots = body->slots.o;
             slots += start;
-            while (i < elems) {
-                MVM_gc_worklist_add(tc, worklist, &slots[i]);
-                i++;
+            MVM_gc_worklist_presize_for(tc, worklist, elems);
+            if (worklist->include_gen2) {
+                for (; i < elems; i++)
+                    MVM_gc_worklist_add_include_gen2_nocheck(tc, worklist, &slots[i]);
+            }
+            else {
+                for (; i < elems; i++)
+                    MVM_gc_worklist_add_no_include_gen2_nocheck(tc, worklist, &slots[i]);
             }
             break;
         }
         case MVM_ARRAY_STR: {
             MVMString **slots = body->slots.s;
             slots += start;
-            while (i < elems) {
-                MVM_gc_worklist_add(tc, worklist, &slots[i]);
-                i++;
+            MVM_gc_worklist_presize_for(tc, worklist, elems);
+            if (worklist->include_gen2) {
+                for (; i < elems; i++)
+                    MVM_gc_worklist_add_include_gen2_nocheck(tc, worklist, &slots[i]);
+            }
+            else {
+                for (; i < elems; i++)
+                    MVM_gc_worklist_add_no_include_gen2_nocheck(tc, worklist, &slots[i]);
             }
             break;
         }
@@ -241,57 +251,48 @@ static void at_pos(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *d
             MVM_exception_throw_adhoc(tc, "MVMArray: Unhandled slot type, got '%s'", MVM_reg_get_debug_name(tc, repr_data->slot_type));
     }
 }
+void MVM_VMArray_at_pos(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *data, MVMint64 index, MVMRegister *value, MVMuint16 kind) {
+    at_pos(tc, st, root, data, index, value, kind);
+}
 
 static MVMuint64 zero_slots(MVMThreadContext *tc, MVMArrayBody *body,
         MVMuint64 elems, MVMuint64 ssize, MVMuint8 slot_type) {
     switch (slot_type) {
         case MVM_ARRAY_OBJ:
-            while (elems < ssize)
-                body->slots.o[elems++] = NULL;
+            memset(&(body->slots.o[elems]), 0, (ssize - elems) * sizeof(MVMObject *));
             break;
         case MVM_ARRAY_STR:
-            while (elems < ssize)
-                body->slots.s[elems++] = NULL;
+            memset(&(body->slots.s[elems]), 0, (ssize - elems) * sizeof(MVMString *));
             break;
         case MVM_ARRAY_I64:
-            while (elems < ssize)
-                body->slots.i64[elems++] = 0;
+            memset(&(body->slots.i64[elems]), 0, (ssize - elems) * sizeof(MVMint64));
             break;
         case MVM_ARRAY_I32:
-            while (elems < ssize)
-                body->slots.i32[elems++] = 0;
+            memset(&(body->slots.i32[elems]), 0, (ssize - elems) * sizeof(MVMint32));
             break;
         case MVM_ARRAY_I16:
-            while (elems < ssize)
-                body->slots.i16[elems++] = 0;
+            memset(&(body->slots.i16[elems]), 0, (ssize - elems) * sizeof(MVMint16));
             break;
         case MVM_ARRAY_I8:
-            while (elems < ssize)
-                body->slots.i8[elems++] = 0;
+            memset(&(body->slots.i8[elems]), 0, (ssize - elems) * sizeof(MVMint8));
             break;
         case MVM_ARRAY_N64:
-            while (elems < ssize)
-                body->slots.n64[elems++] = 0.0;
+            memset(&(body->slots.n64[elems]), 0, (ssize - elems) * sizeof(MVMnum64));
             break;
         case MVM_ARRAY_N32:
-            while (elems < ssize)
-                body->slots.n32[elems++] = 0.0;
+            memset(&(body->slots.n32[elems]), 0, (ssize - elems) * sizeof(MVMnum32));
             break;
         case MVM_ARRAY_U64:
-            while (elems < ssize)
-                body->slots.u64[elems++] = 0;
+            memset(&(body->slots.u64[elems]), 0, (ssize - elems) * sizeof(MVMuint64));
             break;
         case MVM_ARRAY_U32:
-            while (elems < ssize)
-                body->slots.u32[elems++] = 0;
+            memset(&(body->slots.u32[elems]), 0, (ssize - elems) * sizeof(MVMuint32));
             break;
         case MVM_ARRAY_U16:
-            while (elems < ssize)
-                body->slots.u16[elems++] = 0;
+            memset(&(body->slots.u16[elems]), 0, (ssize - elems) * sizeof(MVMuint16));
             break;
         case MVM_ARRAY_U8:
-            while (elems < ssize)
-                body->slots.u8[elems++] = 0;
+            memset(&(body->slots.u8[elems]), 0, (ssize - elems) * sizeof(MVMuint8));
             break;
         default:
             MVM_exception_throw_adhoc(tc, "MVMArray: Unhandled slot type");
@@ -1317,7 +1318,7 @@ static void spesh(MVMThreadContext *tc, MVMSTable *st, MVMSpeshGraph *g, MVMSpes
             ins->operands[0]         = target;
             ins->operands[1].lit_i16 = sizeof(MVMArray);
             ins->operands[2].lit_i16 = MVM_spesh_add_spesh_slot(tc, g, (MVMCollectable *)st);
-            MVM_spesh_get_facts(tc, g, type)->usages--;
+            MVM_spesh_usages_delete_by_reg(tc, g, type, ins);
         }
         break;
     }
@@ -1401,7 +1402,7 @@ static const MVMREPROps VMArray_this_repr = {
     serialize_repr_data,
     deserialize_repr_data,
     deserialize_stable_size,
-    gc_mark,
+    VMArray_gc_mark,
     gc_free,
     NULL, /* gc_cleanup */
     gc_mark_repr_data,
