@@ -1,5 +1,7 @@
 #include "moar.h"
 
+#define LOG_REPOSSESSIONS 0
+
 /* Creates a new serialization context with the specified handle. If any
  * compilation units are waiting for an SC with this handle, removes it from
  * their to-resolve list after installing itself in the appropriate slot. */
@@ -198,7 +200,7 @@ MVMuint8 MVM_sc_is_object_immediately_available(MVMThreadContext *tc, MVMSeriali
 MVMObject * MVM_sc_get_object(MVMThreadContext *tc, MVMSerializationContext *sc, MVMint64 idx) {
     MVMObject **roots = sc->body->root_objects;
     MVMint64    count = sc->body->num_objects;
-    if (idx >= 0 && idx < count)
+    if (MVM_LIKELY(idx >= 0 && idx < count))
         return roots[idx] && !sc_working(sc)
             ? roots[idx]
             : MVM_serialization_demand_object(tc, sc, idx);
@@ -211,14 +213,6 @@ MVMObject * MVM_sc_get_object(MVMThreadContext *tc, MVMSerializationContext *sc,
     }
 }
 
-MVMObject * MVM_sc_get_sc_object(MVMThreadContext *tc, MVMCompUnit *cu,
-                                 MVMuint16 dep, MVMuint64 idx) {
-    MVMSerializationContext *sc = MVM_sc_get_sc(tc, cu, dep);
-    if (sc == NULL)
-        MVM_exception_throw_adhoc(tc, "SC not yet resolved; lookup failed");
-    return MVM_sc_get_object(tc, sc, idx);
-}
-
 /* Given an SC and an index, fetch the object stored there, or return NULL if
  * there is none. Does not cause lazy deserialization. */
 MVMObject * MVM_sc_try_get_object(MVMThreadContext *tc, MVMSerializationContext *sc, MVMint64 idx) {
@@ -255,7 +249,7 @@ void MVM_sc_set_object(MVMThreadContext *tc, MVMSerializationContext *sc, MVMint
 
 /* Given an SC and an index, fetch the STable stored there. */
 MVMSTable * MVM_sc_get_stable(MVMThreadContext *tc, MVMSerializationContext *sc, MVMint64 idx) {
-    if (idx >= 0 && idx < sc->body->num_stables) {
+    if (MVM_LIKELY(idx >= 0 && idx < sc->body->num_stables)) {
         MVMSTable *got = sc->body->root_stables[idx];
         return got && !sc_working(sc) ? got : MVM_serialization_demand_stable(tc, sc, idx);
     }
@@ -279,7 +273,7 @@ MVMSTable * MVM_sc_try_get_stable(MVMThreadContext *tc, MVMSerializationContext
 
 /* Given an SC, an index, and an STable, store the STable at the index. */
 void MVM_sc_set_stable(MVMThreadContext *tc, MVMSerializationContext *sc, MVMint64 idx, MVMSTable *st) {
-    if (idx < 0)
+    if (MVM_UNLIKELY(idx < 0))
         MVM_exception_throw_adhoc(tc,
             "Invalid (negative) STable index %"PRId64, idx);
     if (idx < sc->body->num_stables) {
@@ -398,7 +392,7 @@ void MVM_sc_disclaim(MVMThreadContext *tc, MVMSerializationContext *sc) {
 void MVM_SC_WB_OBJ(MVMThreadContext *tc, MVMObject *obj) {
     assert(!(obj->header.flags & MVM_CF_FORWARDER_VALID));
     assert(MVM_sc_get_idx_of_sc(&obj->header) != ~0);
-    if (MVM_sc_get_idx_of_sc(&obj->header) > 0)
+    if (MVM_sc_get_idx_of_sc(&obj->header) > 0 && !(obj->st->mode_flags & MVM_NEVER_REPOSSESS_TYPE))
         MVM_sc_wb_hit_obj(tc, obj);
 }
 
@@ -458,6 +452,11 @@ void MVM_sc_wb_hit_obj(MVMThreadContext *tc, MVMObject *obj) {
         /* Update SC of the object, claiming it, and update index too. */
         MVM_sc_set_obj_sc(tc, obj, comp_sc);
         MVM_sc_set_idx_in_sc(&(obj->header), new_slot);
+
+#if LOG_REPOSSESSIONS
+        fprintf(stderr, "Repossession of object %s\n", obj->st->debug_name);
+        MVM_dump_backtrace(tc);
+#endif
     }
 }
 
@@ -486,5 +485,10 @@ void MVM_sc_wb_hit_st(MVMThreadContext *tc, MVMSTable *st) {
         /* Update SC of the STable, claiming it. */
         MVM_sc_set_stable_sc(tc, st, comp_sc);
         MVM_sc_set_idx_in_sc(&(st->header), new_slot);
+
+#if LOG_REPOSSESSIONS
+        fprintf(stderr, "Repossession of STable %s\n", st->debug_name);
+        MVM_dump_backtrace(tc);
+#endif
     }
 }
