@@ -101,16 +101,14 @@ static MVMint32 search_frame_handlers_dyn(MVMThreadContext *tc, MVMFrame *f,
                                           LocatedHandler *lh) {
     MVMuint32  i;
     if (f->spesh_cand && f->spesh_cand->jitcode && f->jit_entry_label) {
+        MVMJitCode *jitcode = f->spesh_cand->jitcode;
+        void *current_position = MVM_jit_code_get_current_position(tc, jitcode, f);
         MVMJitHandler    *jhs = f->spesh_cand->jitcode->handlers;
         MVMFrameHandler  *fhs = MVM_frame_effective_handlers(f);
-        MVMint32 num_handlers = f->spesh_cand->jitcode->num_handlers;
-        void         **labels = f->spesh_cand->jitcode->labels;
-        void       *cur_label = f->jit_entry_label;
-        for (i = 0; i < num_handlers; i++) {
-            if (!handler_can_handle(f, &fhs[i], cat, payload))
-                continue;
-            if (cur_label >= labels[jhs[i].start_label] &&
-                cur_label <= labels[jhs[i].end_label] &&
+        for (i = MVM_jit_code_get_active_handlers(tc, jitcode, current_position, 0);
+             i < jitcode->num_handlers;
+             i = MVM_jit_code_get_active_handlers(tc, jitcode, current_position, i+1)) {
+            if (handler_can_handle(f, &fhs[i], cat, payload) &&
                 !in_handler_stack(tc, &fhs[i], f)) {
                 lh->handler     = &fhs[i];
                 lh->jit_handler = &jhs[i];
@@ -178,40 +176,35 @@ static MVMint32 search_frame_handlers_lex(MVMThreadContext *tc, MVMFrame *f,
     MVMuint32 skipping = *skip_first_inlinee;
     MVMFrameHandler *fhs = MVM_frame_effective_handlers(f);
     if (f->spesh_cand && f->spesh_cand->jitcode && f->jit_entry_label) {
-        MVMJitHandler    *jhs = f->spesh_cand->jitcode->handlers;
-        MVMint32 num_handlers = f->spesh_cand->jitcode->num_handlers;
-        void         **labels = f->spesh_cand->jitcode->labels;
-        void       *cur_label = f->jit_entry_label;
-        for (i = 0; i < num_handlers; i++) {
+        MVMJitCode *jitcode = f->spesh_cand->jitcode;
+        void *current_position = MVM_jit_code_get_current_position(tc, jitcode, f);
+        MVMJitHandler    *jhs = jitcode->handlers;
+        for (i = MVM_jit_code_get_active_handlers(tc, jitcode, current_position, 0);
+             i < jitcode->num_handlers;
+             i = MVM_jit_code_get_active_handlers(tc, jitcode, current_position, i+1)) {
             MVMFrameHandler *fh = &(fhs[i]);
             if (skip_all_inlinees && fh->inlinee >= 0)
                 continue;
             if (fh->category_mask == MVM_EX_INLINE_BOUNDARY) {
-                if (cur_label >= labels[jhs[i].start_label] &&
-                        cur_label <= labels[jhs[i].end_label]) {
-                    if (skipping) {
-                        skipping = 0;
-                        *skip_first_inlinee = 0;
+                if (skipping) {
+                    skipping = 0;
+                    *skip_first_inlinee = 0;
+                }
+                else {
+                    MVMuint16 cr_reg = f->spesh_cand->inlines[fh->inlinee].code_ref_reg;
+                    MVMFrame *inline_outer = ((MVMCode *)f->work[cr_reg].o)->body.outer;
+                    if (inline_outer == f) {
+                        skip_all_inlinees = 1;
                     }
                     else {
-                        MVMuint16 cr_reg = f->spesh_cand->inlines[fh->inlinee].code_ref_reg;
-                        MVMFrame *inline_outer = ((MVMCode *)f->work[cr_reg].o)->body.outer;
-                        if (inline_outer == f) {
-                            skip_all_inlinees = 1;
-                        }
-                        else {
-                            *next_outer = inline_outer;
-                            return 0;
-                        }
+                        *next_outer = inline_outer;
+                        return 0;
                     }
                 }
-                continue;
             }
             if (skipping || !handler_can_handle(f, fh, cat, payload))
                 continue;
-            if (cur_label >= labels[jhs[i].start_label] &&
-                    cur_label <= labels[jhs[i].end_label] &&
-                    !in_handler_stack(tc, fh, f)) {
+            if (!in_handler_stack(tc, fh, f)) {
                 if (skipping && f->static_info->body.is_thunk)
                     return 0;
                 lh->handler     = fh;
@@ -281,6 +274,7 @@ static LocatedHandler search_for_handler_from(MVMThreadContext *tc, MVMFrame *f,
     switch (mode) {
         case MVM_EX_THROW_LEX_CALLER:
             skip_first_inlinee = 1;
+            /* fallthrough */
         case MVM_EX_THROW_LEX: {
             MVMint32 skip_all_inlinees = 0;
             while (f != NULL) {
@@ -355,11 +349,11 @@ static void run_handler(MVMThreadContext *tc, LocatedHandler lh, MVMObject *ex_o
         if (lh.jit_handler) {
             void **labels = lh.frame->spesh_cand->jitcode->labels;
             MVMuint8  *pc = lh.frame->spesh_cand->jitcode->bytecode;
-            lh.frame->jit_entry_label = labels[lh.jit_handler->goto_label];
-            MVM_frame_unwind_to(tc, lh.frame, pc, 0, NULL);
+            MVM_frame_unwind_to(tc, lh.frame, pc, 0, NULL, labels[lh.jit_handler->goto_label]);
         } else {
-            MVM_frame_unwind_to(tc, lh.frame, NULL, lh.handler->goto_offset, NULL);
+            MVM_frame_unwind_to(tc, lh.frame, NULL, lh.handler->goto_offset, NULL, NULL);
         }
+
         break;
 
     case MVM_EX_ACTION_INVOKE: {
@@ -411,7 +405,7 @@ static void unwind_after_handler(MVMThreadContext *tc, void *sr_data) {
     MVMException *exception;
     MVMuint32     goto_offset;
     MVMuint8     *abs_address;
-
+    void         *jit_return_label;
 
     /* Get active handler; sanity check (though it's possible other cases
      * should be supported). */
@@ -424,13 +418,14 @@ static void unwind_after_handler(MVMThreadContext *tc, void *sr_data) {
     exception   = (MVMException *)ah->ex_obj;
     if (ah->jit_handler) {
         void **labels = frame->spesh_cand->jitcode->labels;
-        frame->jit_entry_label = labels[ah->jit_handler->goto_label];
+        jit_return_label = labels[ah->jit_handler->goto_label];
         abs_address = frame->spesh_cand->jitcode->bytecode;
         goto_offset = 0;
     }
     else {
         goto_offset = ah->handler->goto_offset;
         abs_address = NULL;
+        jit_return_label = NULL;
     }
     /* Clean up. */
     tc->active_handlers = ah->next_handler;
@@ -438,10 +433,11 @@ static void unwind_after_handler(MVMThreadContext *tc, void *sr_data) {
 
     /* Do the unwinding as needed. */
     if (exception && exception->body.return_after_unwind) {
-        MVM_frame_unwind_to(tc, frame->caller, NULL, 0, tc->last_handler_result);
+        /* we can't very well return to our the unwod JIT address */
+        MVM_frame_unwind_to(tc, frame->caller, NULL, 0, tc->last_handler_result, NULL);
     }
     else {
-        MVM_frame_unwind_to(tc, frame, abs_address, goto_offset, NULL);
+        MVM_frame_unwind_to(tc, frame, abs_address, goto_offset, NULL, jit_return_label);
     }
 }
 
@@ -649,6 +645,9 @@ static void panic_unhandled_cat(MVMThreadContext *tc, MVMuint32 cat) {
 static void panic_unhandled_ex(MVMThreadContext *tc, MVMException *ex) {
     char *backtrace;
 
+    /* If a debug session is running, notify the client. */
+    MVM_debugserver_notify_unhandled_exception(tc, ex);
+
     /* If it's a control exception, try promoting it to a catch one; use
      * the category name. */
     if (ex->body.category != MVM_EX_CAT_CATCH)
@@ -738,10 +737,10 @@ void MVM_exception_throwobj(MVMThreadContext *tc, MVMuint8 mode, MVMObject *ex_o
         ex->body.category = MVM_EX_CAT_CATCH;
     if (resume_result) {
         ex->body.resume_addr = *tc->interp_cur_op;
-        /* Ensure that the jit resume label is stored. The throwish
-         * control guard should ensure that the jit entry label point to
-         * a position just after throwing. */
-        ex->body.jit_resume_label = tc->cur_frame->jit_entry_label;
+        /* Ensure that we store label where the JIT should return, if any */
+        if (tc->jit_return_address != NULL) {
+            ex->body.jit_resume_label = MVM_jit_code_get_current_position(tc, tc->cur_frame->spesh_cand->jitcode, tc->cur_frame);
+        }
     }
     lh = search_for_handler_from(tc, tc->cur_frame, mode, ex->body.category, ex->body.payload);
     if (lh.frame == NULL) {
@@ -810,8 +809,7 @@ void MVM_exception_resume(MVMThreadContext *tc, MVMObject *ex_obj) {
     MVM_free(ah);
 
     /* Unwind to the thrower of the exception; set PC and jit entry label. */
-    target->jit_entry_label = ex->body.jit_resume_label;
-    MVM_frame_unwind_to(tc, target, ex->body.resume_addr, 0, NULL);
+    MVM_frame_unwind_to(tc, target, ex->body.resume_addr, 0, NULL, ex->body.jit_resume_label);
 }
 
 /* Panics and shuts down the VM. Don't do this unless it's something quite
@@ -820,8 +818,7 @@ void MVM_exception_resume(MVMThreadContext *tc, MVMObject *ex_obj) {
  * Use MVM_oops in the case a thread context is available.
  * TODO: Some hook for embedders.
  */
-MVM_NO_RETURN
-void MVM_panic(MVMint32 exitCode, const char *messageFormat, ...) {
+MVM_NO_RETURN void MVM_panic(MVMint32 exitCode, const char *messageFormat, ...) {
     va_list args;
     fprintf(stderr, "MoarVM panic: ");
     va_start(args, messageFormat);
@@ -834,15 +831,13 @@ void MVM_panic(MVMint32 exitCode, const char *messageFormat, ...) {
         exit(exitCode);
 }
 
-MVM_NO_RETURN
-void MVM_panic_allocation_failed(size_t len) {
+MVM_NO_RETURN void MVM_panic_allocation_failed(size_t len) {
     MVM_panic(1, "Memory allocation failed; could not allocate %"MVM_PRSz" bytes", len);
 }
 
 /* A kinder MVM_panic() that doesn't assume our memory is corrupted (but does kill the
  * process to indicate that we've made an error */
-MVM_NO_RETURN
-void MVM_oops(MVMThreadContext *tc, const char *messageFormat, ...) {
+MVM_NO_RETURN void MVM_oops(MVMThreadContext *tc, const char *messageFormat, ...) {
     va_list args;
     fprintf(stderr, "MoarVM oops: ");
     va_start(args, messageFormat);
@@ -857,8 +852,7 @@ void MVM_oops(MVMThreadContext *tc, const char *messageFormat, ...) {
 }
 
 /* Throws an ad-hoc (untyped) exception. */
-MVM_NO_RETURN
-void MVM_exception_throw_adhoc(MVMThreadContext *tc, const char *messageFormat, ...) {
+MVM_NO_RETURN void MVM_exception_throw_adhoc(MVMThreadContext *tc, const char *messageFormat, ...) {
     va_list args;
     va_start(args, messageFormat);
     MVM_exception_throw_adhoc_free_va(tc, NULL, messageFormat, args);
@@ -866,15 +860,13 @@ void MVM_exception_throw_adhoc(MVMThreadContext *tc, const char *messageFormat,
 }
 
 /* Throws an ad-hoc (untyped) exception. */
-MVM_NO_RETURN
-void MVM_exception_throw_adhoc_va(MVMThreadContext *tc, const char *messageFormat, va_list args) {
+MVM_NO_RETURN void MVM_exception_throw_adhoc_va(MVMThreadContext *tc, const char *messageFormat, va_list args) {
     MVM_exception_throw_adhoc_free_va(tc, NULL, messageFormat, args);
 }
 
 /* Throws an ad-hoc (untyped) exception, taking a NULL-terminated array of
  * char pointers to deallocate after message construction. */
-MVM_NO_RETURN
-void MVM_exception_throw_adhoc_free(MVMThreadContext *tc, char **waste, const char *messageFormat, ...) {
+MVM_NO_RETURN void MVM_exception_throw_adhoc_free(MVMThreadContext *tc, char **waste, const char *messageFormat, ...) {
     va_list args;
     va_start(args, messageFormat);
     MVM_exception_throw_adhoc_free_va(tc, waste, messageFormat, args);
@@ -883,8 +875,7 @@ void MVM_exception_throw_adhoc_free(MVMThreadContext *tc, char **waste, const ch
 
 /* Throws an ad-hoc (untyped) exception, taking a NULL-terminated array of
  * char pointers to deallocate after message construction. */
-MVM_NO_RETURN
-void MVM_exception_throw_adhoc_free_va(MVMThreadContext *tc, char **waste, const char *messageFormat, va_list args) {
+MVM_NO_RETURN void MVM_exception_throw_adhoc_free_va(MVMThreadContext *tc, char **waste, const char *messageFormat, va_list args) {
     LocatedHandler lh;
     MVMException *ex;
     /* The current frame will be assigned as the thrower of the exception, so
