@@ -205,6 +205,11 @@ struct MVMInstance {
     /* Fixed size allocator. */
     MVMFixedSizeAlloc *fsa;
 
+    /* Vector of memory to free at the next safepoint, and a mutex to guard
+     * access to it. */
+    MVM_VECTOR_DECL(void *, free_at_safepoint);
+    uv_mutex_t mutex_free_at_safepoint;
+
     /************************************************************************
      * Object system
      ************************************************************************/
@@ -251,6 +256,7 @@ struct MVMInstance {
     /* Flag for if spesh (and certain spesh features) are enabled. */
     MVMint8 spesh_enabled;
     MVMint8 spesh_inline_enabled;
+    MVMint8 spesh_inline_log;
     MVMint8 spesh_osr_enabled;
     MVMint8 spesh_nodelay;
     MVMint8 spesh_blocking;
@@ -352,8 +358,8 @@ struct MVMInstance {
     /* Cache the environment hash */
     MVMObject      *env_hash;
 
-    /* Cache the signal hash */
-    MVMObject      *sig_hash;
+    /* Cache the signal array */
+    MVMObject      *sig_arr;
 
     /* Flags indicating the signals available on the host system */
     MVMuint64       valid_sigs;
@@ -371,7 +377,7 @@ struct MVMInstance {
     uv_mutex_t mutex_int_const_cache;
 
     /* Multi-dispatch cache addition mutex (additions are relatively
-     * rare, so little motivation to have it more fine-grained). */ 
+     * rare, so little motivation to have it more fine-grained). */
     uv_mutex_t mutex_multi_cache_add;
 
     /* Next type cache ID, to go in STable. */
@@ -508,4 +514,8 @@ struct MVMInstance {
 
     /* Flag for if NFA debugging is enabled. */
     MVMint8 nfa_debug_enabled;
+
+    /* Hash Secrets which is used as the hash seed. This is to avoid denial of
+     * service type attacks. */
+    MVMuint64 hashSecrets[2];
 };
