@@ -693,6 +693,8 @@ void MVM_validate_static_frame(MVMThreadContext *tc,
         read_op(val);
         if (val->cur_mark && val->cur_mark[0] == 's')
             fail(val, MSG(val, "Illegal appearance of spesh op"));
+        if (val->cur_info->specializable)
+            fb->specializable = 1;
 
         switch (val->cur_mark[0]) {
             case MARK_regular:
