@@ -14,11 +14,11 @@ void * MVM_gc_allocate_nursery(MVMThreadContext *tc, size_t size) {
      * also; check if we've been signalled to collect. */
     /* Don't use a MVM_load(&tc->gc_status) here for performance, it's okay
      * if the interrupt is delayed a bit. */
-    if (tc->gc_status)
+    if (MVM_UNLIKELY(tc->gc_status))
         MVM_gc_enter_from_interrupt(tc);
 
     /* Guard against 0-byte allocation. */
-    if (size > 0) {
+    if (MVM_LIKELY(size > 0)) {
         /* Do a GC run if this allocation won't fit in what we have
          * left in the nursery. Note this is a loop to handle a
          * pathological case: all the objects in the nursery are very
@@ -26,7 +26,7 @@ void * MVM_gc_allocate_nursery(MVMThreadContext *tc, size_t size) {
          * actually gets freed up. The next run will promote them to the
          * second generation. Note that this circumstance is exceptionally
          * unlikely in any non-contrived situation. */
-        while ((char *)tc->nursery_alloc + size >= (char *)tc->nursery_alloc_limit) {
+        while (MVM_UNLIKELY((char *)tc->nursery_alloc + size >= (char *)tc->nursery_alloc_limit)) {
             if (size > MVM_NURSERY_SIZE)
                 MVM_panic(MVM_exitcode_gcalloc, "Attempt to allocate more than the maximum nursery size");
             MVM_gc_enter_from_allocator(tc);
@@ -88,15 +88,21 @@ MVMObject * MVM_gc_allocate_object(MVMThreadContext *tc, MVMSTable *st) {
         obj->header.size  = (MVMuint16)st->size;
         obj->header.owner = tc->thread_id;
         MVM_ASSIGN_REF(tc, &(obj->header), obj->st, st);
-        if ((obj->header.flags & MVM_CF_SECOND_GEN))
-            if (REPR(obj)->refs_frames)
-                MVM_gc_root_gen2_add(tc, (MVMCollectable *)obj);
         if (st->mode_flags & MVM_FINALIZE_TYPE)
             MVM_gc_finalize_add_to_queue(tc, obj);
     });
     return obj;
 }
 
+/* Allocates a new heap frame. */
+MVMFrame * MVM_gc_allocate_frame(MVMThreadContext *tc) {
+    MVMFrame *f = MVM_gc_allocate_zeroed(tc, sizeof(MVMFrame));
+    f->header.flags |= MVM_CF_FRAME;
+    f->header.size   = sizeof(MVMFrame);
+    f->header.owner  = tc->thread_id;
+    return f;
+}
+
 /* Sets allocate for this thread to be from the second generation by
  * default. */
 void MVM_gc_allocate_gen2_default_set(MVMThreadContext *tc) {
