@@ -65,9 +65,8 @@ void MVM_gc_root_add_permanents_to_worklist(MVMThreadContext *tc, MVMGCWorklist
 /* Adds anything that is a root thanks to being referenced by instance,
  * but that isn't permanent. */
 void MVM_gc_root_add_instance_roots_to_worklist(MVMThreadContext *tc, MVMGCWorklist *worklist, MVMHeapSnapshotState *snapshot) {
-    MVMSerializationContextBody *current, *tmp;
-    MVMLoadedCompUnitName       *current_lcun, *tmp_lcun;
-    unsigned                     bucket_tmp;
+    MVMSerializationContextBody *current;
+    MVMLoadedCompUnitName       *current_lcun;
     MVMString                  **int_to_str_cache;
     MVMuint32                    i;
 
@@ -95,7 +94,7 @@ void MVM_gc_root_add_instance_roots_to_worklist(MVMThreadContext *tc, MVMGCWorkl
 
     /* okay, so this makes the weak hash slightly less weak.. for certain
      * keys of it anyway... */
-    HASH_ITER(hash_handle, tc->instance->sc_weakhash, current, tmp, bucket_tmp) {
+    HASH_ITER_FAST(tc, hash_handle, tc->instance->sc_weakhash, current, {
         /* mark the string handle pointer iff it hasn't yet been resolved */
         add_collectable(tc, worklist, snapshot, current->hash_handle.key,
             "SC weakhash hash key");
@@ -105,14 +104,14 @@ void MVM_gc_root_add_instance_roots_to_worklist(MVMThreadContext *tc, MVMGCWorkl
         else if (!current->claimed)
             add_collectable(tc, worklist, snapshot, current->sc,
                 "SC weakhash unclaimed SC");
-    }
+    });
 
-    HASH_ITER(hash_handle, tc->instance->loaded_compunits, current_lcun, tmp_lcun, bucket_tmp) {
+    HASH_ITER_FAST(tc, hash_handle, tc->instance->loaded_compunits, current_lcun, {
         add_collectable(tc, worklist, snapshot, current_lcun->hash_handle.key,
             "Loaded compilation unit hash key");
         add_collectable(tc, worklist, snapshot, current_lcun->filename,
             "Loaded compilation unit filename");
-    }
+    });
 
     add_collectable(tc, worklist, snapshot, tc->instance->cached_backend_config,
         "Cached backend configuration hash");
@@ -120,8 +119,8 @@ void MVM_gc_root_add_instance_roots_to_worklist(MVMThreadContext *tc, MVMGCWorkl
     add_collectable(tc, worklist, snapshot, tc->instance->env_hash,
         "Cached environment variable hash");
 
-    add_collectable(tc, worklist, snapshot, tc->instance->sig_hash,
-        "Cached signal mapping hash");
+    add_collectable(tc, worklist, snapshot, tc->instance->sig_arr,
+        "Cached signal mapping array");
 
     MVM_debugserver_mark_handles(tc, worklist, snapshot);
 }
@@ -129,8 +128,7 @@ void MVM_gc_root_add_instance_roots_to_worklist(MVMThreadContext *tc, MVMGCWorkl
 /* Adds anything that is a root thanks to being referenced by a thread,
  * context, but that isn't permanent. */
 void MVM_gc_root_add_tc_roots_to_worklist(MVMThreadContext *tc, MVMGCWorklist *worklist, MVMHeapSnapshotState *snapshot) {
-    MVMNativeCallbackCacheHead *current_cbceh, *tmp_cbceh;
-    unsigned bucket_tmp;
+    MVMNativeCallbackCacheHead *current_cbceh;
 
     /* Any active exception handlers and payload. */
     MVMActiveHandler *cur_ah = tc->active_handlers;
@@ -164,7 +162,7 @@ void MVM_gc_root_add_tc_roots_to_worklist(MVMThreadContext *tc, MVMGCWorklist *w
     add_collectable(tc, worklist, snapshot, tc->cur_dispatcher_for, "Current dispatcher for");
 
     /* Callback cache. */
-    HASH_ITER(hash_handle, tc->native_callback_cache, current_cbceh, tmp_cbceh, bucket_tmp) {
+    HASH_ITER_FAST(tc, hash_handle, tc->native_callback_cache, current_cbceh, {
         MVMint32 i;
         MVMNativeCallback *entry = current_cbceh->head;
         add_collectable(tc, worklist, snapshot, current_cbceh->hash_handle.key,
@@ -177,7 +175,7 @@ void MVM_gc_root_add_tc_roots_to_worklist(MVMThreadContext *tc, MVMGCWorklist *w
                 "Native callback cache target");
             entry = entry->next;
         }
-    }
+    });
 
     /* Profiling data. */
     if (worklist)
@@ -187,13 +185,17 @@ void MVM_gc_root_add_tc_roots_to_worklist(MVMThreadContext *tc, MVMGCWorklist *w
     add_collectable(tc, worklist, snapshot, tc->serialized_string_heap,
         "Serialized string heap");
 
-    /* Specialization log and stack simulation. */
+    /* Specialization log, stack simulation, and plugin state. */
     add_collectable(tc, worklist, snapshot, tc->spesh_log, "Specialization log");
     if (worklist)
         MVM_spesh_sim_stack_gc_mark(tc, tc->spesh_sim_stack, worklist);
-    else {
+    else
         MVM_spesh_sim_stack_gc_describe(tc, snapshot, tc->spesh_sim_stack);
-    }
+    if (tc->spesh_active_graph)
+        MVM_spesh_graph_mark(tc, tc->spesh_active_graph, worklist);
+    MVM_spesh_plugin_guard_list_mark(tc, tc->plugin_guards, tc->num_plugin_guards, worklist);
+    add_collectable(tc, worklist, snapshot, tc->plugin_guard_args,
+        "Plugin guard args");
 
     if (tc->step_mode_frame)
         add_collectable(tc, worklist, snapshot, tc->step_mode_frame,
@@ -272,7 +274,7 @@ void MVM_gc_root_gen2_add(MVMThreadContext *tc, MVMCollectable *c) {
     if (tc->num_gen2roots == tc->alloc_gen2roots) {
         tc->alloc_gen2roots *= 2;
         tc->gen2roots = MVM_realloc(tc->gen2roots,
-            sizeof(MVMCollectable **) * tc->alloc_gen2roots);
+            sizeof(MVMCollectable *) * tc->alloc_gen2roots);
     }
 
     /* Add this one to the list. */
@@ -392,6 +394,7 @@ void MVM_gc_root_add_frame_roots_to_worklist(MVMThreadContext *tc, MVMGCWorklist
         MVM_gc_worklist_add(tc, worklist, &e->invoked_call_capture);
         if (e->dynlex_cache_name)
             MVM_gc_worklist_add(tc, worklist, &e->dynlex_cache_name);
+        MVM_gc_worklist_add(tc, worklist, &e->exit_handler_result);
     }
 
     /* Scan the registers. */
