@@ -340,8 +340,7 @@ MVMint32 values_cmp_last_ref(LiveRange *values, MVMint32 a, MVMint32 b) {
 }
 
 /* register assignment logic */
-#define ARRAY_SIZE(a) (sizeof(a)/sizeof(a[0]))
-#define NEXT_IN_RING(a,x) (((x)+1) == ARRAY_SIZE(a) ? 0 : ((x)+1))
+#define NEXT_IN_RING(a,x) (((x)+1) == MVM_ARRAY_SIZE(a) ? 0 : ((x)+1))
 MVMint8 get_register(MVMThreadContext *tc, RegisterAllocator *alc, MVMJitStorageClass reg_cls) {
     /* ignore storage class for now */
     MVMint8 reg_num;
@@ -422,6 +421,7 @@ MVM_STATIC_INLINE void open_hole(RegisterAllocator *alc, MVMint32 ref, MVMint32
 static void find_holes(MVMThreadContext *tc, RegisterAllocator *alc, MVMJitTileList *list) {
     MVMint32 i, j, k;
 
+    MVMJitExprTree *tree = list->tree;
     MVMint32 bitmap_size = (alc->values_num >> 5) + 1;
 
  /* convenience macros */
@@ -461,10 +461,10 @@ static void find_holes(MVMThreadContext *tc, RegisterAllocator *alc, MVMJitTileL
             MVMJitTile *tile = list->items[i];
             if (tile->op == MVM_JIT_ARGLIST) {
                 /* list of uses, all very real */
-                MVMint32 nchild = list->tree->nodes[tile->node + 1];
+                MVMint32 nchild = MVM_JIT_EXPR_NCHILD(tree, tile->node);
+                MVMint32 *args  = MVM_JIT_EXPR_LINKS(tree, tile->node);
                 for (k = 0; k < nchild; k++) {
-                    MVMint32 carg = list->tree->nodes[tile->node + 2 + k];
-                    MVMint32 ref  = value_set_find(alc->sets, list->tree->nodes[carg + 1])->idx;
+                    MVMint32 ref  = value_set_find(alc->sets, MVM_JIT_EXPR_LINKS(tree, args[k])[0])->idx;
                     if (!MVM_bitmap_get(live_in, ref)) {
                         MVM_bitmap_set(live_in, ref);
                         close_hole(alc, ref, i);
@@ -502,9 +502,9 @@ static void find_holes(MVMThreadContext *tc, RegisterAllocator *alc, MVMJitTileL
 
 
 static void determine_live_ranges(MVMThreadContext *tc, RegisterAllocator *alc, MVMJitTileList *list) {
-    MVMJitExprTree *tree = list->tree;
     MVMint32 i, j, n;
     MVMint32 num_phi = 0; /* pessimistic but correct upper bound of number of holes */
+    MVMJitExprTree *tree = list->tree;
 
     alc->sets = MVM_calloc(tree->nodes_num, sizeof(UnionFind));
     /* up to 4 refs per tile (1 out, 3 in) plus the number of refs per arglist */
@@ -520,27 +520,30 @@ static void determine_live_ranges(MVMThreadContext *tc, RegisterAllocator *alc,
         /* Each of the following counts as either an alias or as a PHI (in case
          * of IF), and thus these are not actual definitions */
         if (tile->op == MVM_JIT_COPY) {
-            MVMint32 ref        = tree->nodes[tile->node + 1];
+            MVMint32 ref        = MVM_JIT_EXPR_LINKS(tree, node)[0];
+            _DEBUG("Unify COPY node (%d -> %d)", tile->node, ref);
             alc->sets[node].key = ref; /* point directly to actual definition */
         } else if (tile->op == MVM_JIT_DO) {
-            MVMint32 nchild     = tree->nodes[tile->node + 1];
-            MVMint32 ref        = tree->nodes[tile->node + 1 + nchild];
+            MVMint32 last_child = MVM_JIT_EXPR_FIRST_CHILD(tree, node) + MVM_JIT_EXPR_NCHILD(tree, node) - 1;
+            MVMint32 ref        = tree->nodes[last_child];
+            _DEBUG("Unify COPY DO (%d -> %d)", tile->node, ref);
             alc->sets[node].key = ref;
         } else if (tile->op == MVM_JIT_IF) {
-            MVMint32 left_cond   = tree->nodes[tile->node + 2];
-            MVMint32 right_cond  = tree->nodes[tile->node + 3];
+            MVMint32 *links     = MVM_JIT_EXPR_LINKS(tree, node);
             /* NB; this may cause a conflict in register requirements, in which
              * case we should resolve it by creating a new live range or inserting
              * a copy */
-            alc->sets[node].key  = value_set_union(alc->sets, alc->values, left_cond, right_cond);
+            alc->sets[node].key  = value_set_union(alc->sets, alc->values, links[1], links[2]);
+            _DEBUG("Merging nodes %d and %d to %d (result key = %d)", links[1], links[2], node, alc->sets[node].key);
             num_phi++;
         } else if (tile->op == MVM_JIT_ARGLIST) {
-            MVMint32 num_args = tree->nodes[tile->node + 1];
-            MVMJitExprNode *refs = tree->nodes + tile->node + 2;
+            MVMint32 num_args = MVM_JIT_EXPR_NCHILD(tree, node);
+            MVMint32 *refs = MVM_JIT_EXPR_LINKS(tree, node);
+            _DEBUG("Adding %d references to ARGLIST node", num_args);
             for (j = 0; j < num_args; j++) {
-                MVMint32 carg  = refs[j];
-                MVMint32 value = list->tree->nodes[carg+1];
+                MVMint32 value = MVM_JIT_EXPR_LINKS(tree, refs[j])[0];
                 MVMint32 idx   = value_set_find(alc->sets, value)->idx;
+                _DEBUG("  Reference %d", idx);
                 live_range_add_ref(alc, alc->values + idx, i, j + 1);
                 /* include the CALL node into the arglist child range, so we
                  * don't release them too early */
@@ -553,6 +556,7 @@ static void determine_live_ranges(MVMThreadContext *tc, RegisterAllocator *alc,
                 MVMint32 idx          = live_range_init(alc);
                 alc->sets[node].key   = node;
                 alc->sets[node].idx   = idx;
+                _DEBUG("Create live range %d (tile=%d, node=%d)", idx,i, node);
                 live_range_add_ref(alc, alc->values + idx, i, 0);
                 if (MVM_JIT_REGISTER_HAS_REQUIREMENT(register_spec)) {
                     alc->values[idx].register_spec = register_spec;
@@ -565,17 +569,18 @@ static void determine_live_ranges(MVMThreadContext *tc, RegisterAllocator *alc,
                 /* any 'use' register requirements are handled in the allocation step */
                 if (MVM_JIT_REGISTER_IS_USED(register_spec)) {
                     MVMint32 idx = value_set_find(alc->sets, tile->refs[j])->idx;
+                    _DEBUG("Adding reference to live range %d from tile %d", idx, i);
                     live_range_add_ref(alc, alc->values + idx, i, j + 1);
                 }
             }
         }
-        if (MVM_JIT_TILE_YIELDS_VALUE(tile) && tree->info[node].opr_type != 0) {
+        if (MVM_JIT_TILE_YIELDS_VALUE(tile) && MVM_JIT_EXPR_INFO(tree, node)->type != 0) {
             LiveRange *range = alc->values + value_set_find(alc->sets, node)->idx;
-            _ASSERT(range->reg_type == 0 || (range->reg_type << 3) == tree->info[node].opr_type,
+            _ASSERT(range->reg_type == 0 || (range->reg_type) == MVM_JIT_EXPR_INFO(tree, node)->type,
                     "Register types do not match between value and node");
             /* shift to match MVM_reg_types. should arguably be a macro maybe */
-            range->reg_type = tree->info[node].opr_type >> 3;
-            _DEBUG( "Assigned type: %d\n", range->reg_type);
+            range->reg_type = MVM_JIT_EXPR_INFO(tree, node)->type;
+            _DEBUG( "Assigned type: %d to live range %d\n", range->reg_type, range - alc->values);
         }
     }
     if (num_phi > 0) {
@@ -637,7 +642,7 @@ static void active_set_expire(MVMThreadContext *tc, RegisterAllocator *alc, MVMi
             break;
         } else {
             _DEBUG("Live range %d is out of scope (last ref %d, %d) and releasing register %d",
-                    v, values[v].end, order_nr, reg_num);
+                   v, alc->values[v].end, order_nr, reg_num);
             free_register(tc, alc, MVM_JIT_STORAGE_GPR, reg_num);
         }
     }
@@ -652,9 +657,19 @@ static void active_set_expire(MVMThreadContext *tc, RegisterAllocator *alc, MVMi
     }
 }
 
+/* Compute the earliest live range that is still active. */
+static MVMint32 earliest_active_value(MVMThreadContext *tc, RegisterAllocator *alc, MVMint32 tile_order_nr) {
+    /* can we cache this, and does it make sense to do so? */
+    int i;
+    for (i = 0; i < alc->active_top; i++) {
+        tile_order_nr = MIN(tile_order_nr, alc->values[alc->active[i]].start);
+    }
+    return tile_order_nr;
+}
+
 static void active_set_splice(MVMThreadContext *tc, RegisterAllocator *alc, MVMint32 to_splice) {
     MVMint32 i ;
-    /* find (reverse, because it's usually the last); predecrement alc->active_top */
+    /* find (reverse, because it's usually the last); predecrement alc->active_top because we're removing one item */
     for (i = --alc->active_top; i >= 0; i--) {
         if (alc->active[i] == to_splice)
             break;
@@ -712,12 +727,12 @@ static MVMint32 select_live_range_for_spill(MVMThreadContext *tc, RegisterAlloca
 }
 
 
-
 static void live_range_spill(MVMThreadContext *tc, RegisterAllocator *alc, MVMJitTileList *list,
                              MVMint32 to_spill, MVMint32 spill_pos, MVMint32 code_pos) {
 
     MVMint8 reg_spilled = alc->values[to_spill].reg_num;
     /* loop over all value refs */
+    _DEBUG("Spilling live range value %d to memory position %d at %d", to_spill, spill_pos, code_pos);
 
     while (alc->values[to_spill].first != NULL) {
         /* make a new live range */
@@ -771,7 +786,8 @@ static void prepare_arglist_and_call(MVMThreadContext *tc, RegisterAllocator *al
     MVMJitTile *arglist_tile = list->items[arglist_idx],
                   *call_tile = list->items[call_idx];
     MVMJitExprTree *tree = list->tree;
-    MVMint32 num_args = tree->nodes[arglist_tile->node + 1];
+    MVMint32 num_args = MVM_JIT_EXPR_NCHILD(tree, arglist_tile->node);
+    MVMint32 *args    = MVM_JIT_EXPR_LINKS(tree, arglist_tile->node);
     MVMint32           arg_values[16];
     MVMJitStorageRef storage_refs[16];
 
@@ -803,15 +819,14 @@ static void prepare_arglist_and_call(MVMThreadContext *tc, RegisterAllocator *al
 
     /* get value refs for arglist */
     for (i = 0; i < num_args; i++) {
-        MVMint32 carg  = tree->nodes[arglist_tile->node + 2 + i];
         /* may refer to spilled live range */
-        arg_values[i] = value_set_find(alc->sets, tree->nodes[carg + 1])->idx;
+        arg_values[i] = value_set_find(alc->sets, MVM_JIT_EXPR_LINKS(tree, args[i])[0])->idx;
     }
 
     _DEBUG("prepare_call: Got %d args", num_args);
 
     /* initialize topological map, use -1 as 'undefined' inboud value */
-    for (i = 0; i < ARRAY_SIZE(topological_map); i++) {
+    for (i = 0; i < MVM_ARRAY_SIZE(topological_map); i++) {
         topological_map[i].num_out =  0;
         topological_map[i].in_reg  = -1;
     }
@@ -925,7 +940,7 @@ static void prepare_arglist_and_call(MVMThreadContext *tc, RegisterAllocator *al
 
     /* at this point, all outbound edges have been created, and none have been
      * processed yet, so we can eqnueue all 'free' transfers */
-    for (i = 0; i < ARRAY_SIZE(topological_map); i++) {
+    for (i = 0; i < MVM_ARRAY_SIZE(topological_map); i++) {
         if (topological_map[i].num_out == 0 &&
             topological_map[i].in_reg >= 0) {
             _DEBUG("Directly transfer %d -> %d", topological_map[i].in_reg, i);
@@ -989,6 +1004,7 @@ static void prepare_arglist_and_call(MVMThreadContext *tc, RegisterAllocator *al
         MVMint32 arg = spilled_args[i];
         LiveRange *v = alc->values + arg_values[arg];
         if (storage_refs[arg]._cls == MVM_JIT_STORAGE_GPR) {
+            _DEBUG("Loading spilled value to Rq(%d) from [rbx+%d]", storage_refs[arg]._pos, v->spill_pos);
             INSERT_LOAD_LOCAL(storage_refs[arg]._pos, v->spill_pos);
         } else if (storage_refs[arg]._cls == MVM_JIT_STORAGE_STACK) {
             INSERT_LOCAL_STACK_COPY(storage_refs[arg]._pos, v->spill_pos);
@@ -1049,7 +1065,7 @@ static void process_live_range(MVMThreadContext *tc, RegisterAllocator *alc, MVM
     MVMint32 tile_order_nr = alc->values[v].start;
     if (MVM_JIT_REGISTER_HAS_REQUIREMENT(alc->values[v].register_spec)) {
         reg = MVM_JIT_REGISTER_REQUIREMENT(alc->values[v].register_spec);
-        if (MVM_bitmap_get((MVMBitmap*)&NVR_GPR_BITMAP, reg)) {
+        if (MVM_bitmap_get_low(NVR_GPR_BITMAP, reg)) {
             assign_register(tc, alc, list, v, MVM_JIT_STORAGE_NVR, reg);
         } else {
             /* TODO; might require swapping / spilling */
@@ -1086,7 +1102,14 @@ static void linear_scan(MVMThreadContext *tc, RegisterAllocator *alc, MVMJitTile
             _DEBUG("Processing live range %d (first ref %d, last ref %d)",
                    live_range, alc->values[live_range].start, alc->values[live_range].end);
             active_set_expire(tc, alc, tile_order_nr);
-            spill_set_release(tc, alc, tile_order_nr);
+            /* We can release the spill slot only if there is no more active
+             * live range overlapping with its extent. Otherwise, when we reuse
+             * the slot, we risk overwriting a useful value.
+             *
+             * We pass the current tile_order_nr as the upper bound (e.g. when
+             * there may be no active live ranges, slots will still be useful if
+             * they have later uses */
+            spill_set_release(tc, alc,  earliest_active_value(tc,alc, tile_order_nr));
             process_live_range(tc, alc, list, live_range);
         } else {
             /* still have tiles to process, increment cursor */
