@@ -1,10 +1,31 @@
 #include "moar.h"
 #include <math.h>
 
+#ifndef MANTISSA_BITS_IN_DOUBLE
+#define MANTISSA_BITS_IN_DOUBLE 53
+#endif
+#ifndef MAX_BIGINT_BITS_IN_DOUBLE
+#define MAX_BIGINT_BITS_IN_DOUBLE 1023
+#endif
+
 #ifndef MAX
     #define MAX(x,y) ((x)>(y)?(x):(y))
 #endif
 
+#ifndef MIN
+    #define MIN(x,y) ((x)<(y)?(x):(y))
+#endif
+
+MVM_STATIC_INLINE void adjust_nursery(MVMThreadContext *tc, MVMP6bigintBody *body) {
+    if (MVM_BIGINT_IS_BIG(body)) {
+        int used = USED(body->u.bigint);
+        int adjustment = MIN(used, 32768) & ~0x7;
+        if (adjustment && (char *)tc->nursery_alloc_limit - adjustment > (char *)tc->nursery_alloc) {
+            tc->nursery_alloc_limit = (char *)(tc->nursery_alloc_limit) - adjustment;
+        }
+    }
+}
+
 /* Taken from mp_set_long, but portably accepts a 64-bit number. */
 int MVM_bigint_mp_set_uint64(mp_int * a, MVMuint64 b) {
   int     x, res;
@@ -31,33 +52,47 @@ int MVM_bigint_mp_set_uint64(mp_int * a, MVMuint64 b) {
   return MP_OKAY;
 }
 
-static MVMnum64 mp_get_double(mp_int *a) {
-    MVMnum64 d    = 0.0;
-    MVMnum64 sign = SIGN(a) == MP_NEG ? -1.0 : 1.0;
-    int i;
-    if (USED(a) == 0)
-        return d;
-    if (USED(a) == 1)
-        return sign * (MVMnum64) DIGIT(a, 0);
+/*
+ *  Convert to double, assumes IEEE-754 conforming double. Taken from
+ *  https://github.com/czurnieden/libtommath/blob/master/bn_mp_get_double.c
+ *  and slightly modified to fit MoarVM's setup.
+ */
+static const int mp_get_double_digits_needed
+= ((MANTISSA_BITS_IN_DOUBLE + DIGIT_BIT) / DIGIT_BIT) + 1;
+static const double mp_get_double_multiplier = (double)(MP_MASK + 1);
+
+static MVMnum64 mp_get_double(mp_int *a, int shift) {
+    MVMnum64 d;
+    int i, limit, final_shift;
+    d = 0.0;
 
     mp_clamp(a);
-    i = USED(a) - 1;
-    d = (MVMnum64) DIGIT(a, i);
-    i--;
-    if (i == -1) {
-        return sign * d;
-    }
-    d *= pow(2.0, DIGIT_BIT);
-    d += (MVMnum64) DIGIT(a, i);
+    i = a->used;
+    limit = (i <= mp_get_double_digits_needed)
+        ? 0 : i - mp_get_double_digits_needed;
 
-    if (USED(a) > 2) {
-        i--;
-        d *= pow(2.0, DIGIT_BIT);
-        d += (MVMnum64) DIGIT(a, i);
+    while (i-- > limit) {
+        d += a->dp[i];
+        d *= mp_get_double_multiplier;
     }
 
-    d *= pow(2.0, DIGIT_BIT * i);
-    return sign * d;
+    if (a->sign == MP_NEG)
+        d *= -1.0;
+    final_shift = i * DIGIT_BIT - shift;
+    if (final_shift < 0) {
+        while (final_shift < -1023) {
+            d *= pow(2.0, -1023);
+            final_shift += 1023;
+        }
+    }
+    else {
+        while (final_shift > 1023) {
+            d *= pow(2.0, 1023);
+            final_shift -= 1023;
+        }
+    }
+    d *= pow(2.0, final_shift);
+    return d;
 }
 
 static void from_num(MVMnum64 d, mp_int *a) {
@@ -131,7 +166,7 @@ static mp_int * force_bigint(const MVMP6bigintBody *body, mp_int **tmp) {
         return body->u.bigint;
     }
     else {
-        MVMint32 value = body->u.smallint.value;
+        MVMint64 value = body->u.smallint.value;
         mp_int *i = MVM_malloc(sizeof(mp_int));
         mp_init(i);
         if (value >= 0) {
@@ -179,13 +214,13 @@ static void store_int64_result(MVMP6bigintBody *body, MVMint64 result) {
     }
 }
 
-/* Stores an bigint in a bigint result body, either as a 32-bit smallint if it
+/* Stores a bigint in a bigint result body, either as a 32-bit smallint if it
  * is in range, or a big integer if not. Clears and frees the passed bigint if
  * it is not being used. */
 static void store_bigint_result(MVMP6bigintBody *body, mp_int *i) {
     if (can_be_smallint(i)) {
         body->u.smallint.flag = MVM_BIGINT_32_FLAG;
-        body->u.smallint.value = SIGN(i) ? -DIGIT(i, 0) : DIGIT(i, 0);
+        body->u.smallint.value = SIGN(i) == MP_NEG ? -DIGIT(i, 0) : DIGIT(i, 0);
         mp_clear(i);
         MVM_free(i);
     }
@@ -194,12 +229,12 @@ static void store_bigint_result(MVMP6bigintBody *body, mp_int *i) {
     }
 }
 
-/* Bitops on libtomath (no 2s compliment API) are horrendously inefficient and
+/* Bitops on libtomath (no two's complement API) are horrendously inefficient and
  * really should be hand-coded to work DIGIT-by-DIGIT with in-loop carry
  * handling.  For now we have these fixups.
  *
  * The following inverts the bits of a negative bigint, adds 1 to that, and
- * appends sign-bit extension DIGITs to it to give us a 2s compliment
+ * appends sign-bit extension DIGITs to it to give us a 2s complement
  * representation in memory.  Do not call it on positive bigints.
  */
 static void grow_and_negate(const mp_int *a, int size, mp_int *b) {
@@ -226,7 +261,7 @@ static void grow_and_negate(const mp_int *a, int size, mp_int *b) {
 }
 
 static void two_complement_bitop(mp_int *a, mp_int *b, mp_int *c,
-				 int (*mp_bitop)(mp_int *, mp_int *, mp_int *)) {
+                                 int (*mp_bitop)(mp_int *, mp_int *, mp_int *)) {
 
     mp_int d;
     mp_int e;
@@ -246,7 +281,7 @@ static void two_complement_bitop(mp_int *a, mp_int *b, mp_int *c,
         g = &e;
     }
     /* f and g now guaranteed to each point to positive bigints containing
-     * a 2s compliment representation of the values in a and b.  If either
+     * a two's complement representation of the values in a and b.  If either
      * a or b was negative, the representation is one tomath "digit" longer
      * than it need be and sign extended.
      */
@@ -266,7 +301,7 @@ static void two_complement_bitop(mp_int *a, mp_int *b, mp_int *c,
             DIGIT(c, i) = (~DIGIT(c, i)) & MP_MASK;
         }
         mp_add_d(c, 1, c);
-	mp_neg(c, c);
+        mp_neg(c, c);
     }
 }
 
@@ -288,8 +323,13 @@ static void two_complement_shl(mp_int *result, mp_int *value, MVMint64 count) {
 }
 
 #define MVM_BIGINT_UNARY_OP(opname, SMALLINT_OP) \
-void MVM_bigint_##opname(MVMThreadContext *tc, MVMObject *result, MVMObject *source) { \
-    MVMP6bigintBody *bb = get_bigint_body(tc, result); \
+MVMObject * MVM_bigint_##opname(MVMThreadContext *tc, MVMObject *result_type, MVMObject *source) { \
+    MVMP6bigintBody *bb; \
+    MVMObject *result; \
+    MVMROOT(tc, source, { \
+        result = MVM_repr_alloc_init(tc, result_type);\
+    }); \
+    bb = get_bigint_body(tc, result); \
     if (!IS_CONCRETE(source)) { \
         store_int64_result(bb, 0); \
     } \
@@ -301,6 +341,7 @@ void MVM_bigint_##opname(MVMThreadContext *tc, MVMObject *result, MVMObject *sou
             mp_init(ib); \
             mp_##opname(ia, ib); \
             store_bigint_result(bb, ib); \
+            adjust_nursery(tc, bb); \
         } \
         else { \
             MVMint64 sb; \
@@ -309,6 +350,7 @@ void MVM_bigint_##opname(MVMThreadContext *tc, MVMObject *result, MVMObject *sou
             store_int64_result(bb, sb); \
         } \
     } \
+    return result; \
 }
 
 #define MVM_BIGINT_BINARY_OP(opname) \
@@ -317,11 +359,9 @@ MVMObject * MVM_bigint_##opname(MVMThreadContext *tc, MVMObject *result_type, MV
     MVMObject *result; \
     mp_int *tmp[2] = { NULL, NULL }; \
     mp_int *ia, *ib, *ic; \
-    MVMROOT(tc, a, { \
-    MVMROOT(tc, b, { \
+    MVMROOT2(tc, a, b, { \
         result = MVM_repr_alloc_init(tc, result_type);\
     }); \
-    }); \
     ba = get_bigint_body(tc, a); \
     bb = get_bigint_body(tc, b); \
     bc = get_bigint_body(tc, result); \
@@ -332,6 +372,7 @@ MVMObject * MVM_bigint_##opname(MVMThreadContext *tc, MVMObject *result_type, MV
     mp_##opname(ia, ib, ic); \
     store_bigint_result(bc, ic); \
     clear_temp_bigints(tmp, 2); \
+    adjust_nursery(tc, bc); \
     return result; \
 }
 
@@ -344,11 +385,9 @@ MVMObject * MVM_bigint_##opname(MVMThreadContext *tc, MVMObject *result_type, MV
     if (MVM_BIGINT_IS_BIG(ba) || MVM_BIGINT_IS_BIG(bb)) { \
         mp_int *tmp[2] = { NULL, NULL }; \
         mp_int *ia, *ib, *ic; \
-        MVMROOT(tc, a, { \
-        MVMROOT(tc, b, { \
+        MVMROOT2(tc, a, b, { \
             result = MVM_repr_alloc_init(tc, result_type);\
         }); \
-        }); \
         ba = get_bigint_body(tc, a); \
         bb = get_bigint_body(tc, b); \
         bc = get_bigint_body(tc, result); \
@@ -359,6 +398,7 @@ MVMObject * MVM_bigint_##opname(MVMThreadContext *tc, MVMObject *result_type, MV
         mp_##opname(ia, ib, ic); \
         store_bigint_result(bc, ic); \
         clear_temp_bigints(tmp, 2); \
+        adjust_nursery(tc, bc); \
     } \
     else { \
         MVMint64 sc; \
@@ -381,11 +421,9 @@ MVMObject * MVM_bigint_##opname(MVMThreadContext *tc, MVMObject *result_type, MV
     MVMP6bigintBody *bb = get_bigint_body(tc, b); \
     MVMP6bigintBody *bc; \
     MVMObject *result; \
-    MVMROOT(tc, a, { \
-    MVMROOT(tc, b, { \
+    MVMROOT2(tc, a, b, { \
         result = MVM_repr_alloc_init(tc, result_type);\
     }); \
-    }); \
     bc = get_bigint_body(tc, result); \
     if (MVM_BIGINT_IS_BIG(ba) || MVM_BIGINT_IS_BIG(bb)) { \
         mp_int *tmp[2] = { NULL, NULL }; \
@@ -396,6 +434,7 @@ MVMObject * MVM_bigint_##opname(MVMThreadContext *tc, MVMObject *result_type, MV
         two_complement_bitop(ia, ib, ic, mp_##opname); \
         store_bigint_result(bc, ic); \
         clear_temp_bigints(tmp, 2); \
+        adjust_nursery(tc, bc); \
     } \
     else { \
         MVMint64 sc; \
@@ -424,11 +463,9 @@ MVMObject *MVM_bigint_gcd(MVMThreadContext *tc, MVMObject *result_type, MVMObjec
     MVMP6bigintBody *bc;
     MVMObject       *result;
 
-    MVMROOT(tc, a, {
-    MVMROOT(tc, b, {
+    MVMROOT2(tc, a, b, {
         result = MVM_repr_alloc_init(tc, result_type);
     });
-    });
 
     bc = get_bigint_body(tc, result);
 
@@ -441,6 +478,7 @@ MVMObject *MVM_bigint_gcd(MVMThreadContext *tc, MVMObject *result_type, MVMObjec
         mp_gcd(ia, ib, ic);
         store_bigint_result(bc, ic);
         clear_temp_bigints(tmp, 2);
+        adjust_nursery(tc, bc);
     } else {
         MVMint32 sa = ba->u.smallint.value;
         MVMint32 sb = bb->u.smallint.value;
@@ -487,17 +525,15 @@ MVMObject * MVM_bigint_mod(MVMThreadContext *tc, MVMObject *result_type, MVMObje
 
     MVMObject *result;
 
-    MVMROOT(tc, a, {
-    MVMROOT(tc, b, {
+    MVMROOT2(tc, a, b, {
         result = MVM_repr_alloc_init(tc, result_type);
     });
-    });
 
     bc = get_bigint_body(tc, result);
 
-    // XXX the behavior of C's mod operator is not correct
-    // for our purposes. So we rely on mp_mod for all our modulus
-    // calculations for now.
+    /* XXX the behavior of C's mod operator is not correct
+     * for our purposes. So we rely on mp_mod for all our modulus
+     * calculations for now. */
     if (1 || MVM_BIGINT_IS_BIG(ba) || MVM_BIGINT_IS_BIG(bb)) {
         mp_int *tmp[2] = { NULL, NULL };
         mp_int *ia = force_bigint(ba, tmp);
@@ -514,6 +550,7 @@ MVMObject * MVM_bigint_mod(MVMThreadContext *tc, MVMObject *result_type, MVMObje
             MVM_exception_throw_adhoc(tc, "Division by zero");
         }
         store_bigint_result(bc, ic);
+        adjust_nursery(tc, bc);
     } else {
         store_int64_result(bc, ba->u.smallint.value % bb->u.smallint.value);
     }
@@ -534,22 +571,24 @@ MVMObject *MVM_bigint_div(MVMThreadContext *tc, MVMObject *result_type, MVMObjec
 
     int mp_result;
 
-    MVMROOT(tc, a, {
-    MVMROOT(tc, b, {
+    if (!MVM_BIGINT_IS_BIG(bb) && bb->u.smallint.value == 1 && STABLE(a) == STABLE(b)) {
+        return a;
+    }
+
+    MVMROOT2(tc, a, b, {
         result = MVM_repr_alloc_init(tc, result_type);
     });
-    });
 
     bc = get_bigint_body(tc, result);
 
+    /* we only care about MP_LT or !MP_LT, so we give MP_GT even for 0. */
     if (MVM_BIGINT_IS_BIG(ba)) {
-        cmp_a = mp_cmp_d(ba->u.bigint, 0);
+        cmp_a = !mp_iszero(ba->u.bigint) && SIGN(ba->u.bigint) == MP_NEG ? MP_LT : MP_GT;
     } else {
-        // we only care about MP_LT or !MP_LT, so we give MP_GT even for 0.
         cmp_a = ba->u.smallint.value < 0 ? MP_LT : MP_GT;
     }
     if (MVM_BIGINT_IS_BIG(bb)) {
-        cmp_b = mp_cmp_d(bb->u.bigint, 0);
+        cmp_b = !mp_iszero(bb->u.bigint) && SIGN(bb->u.bigint) == MP_NEG ? MP_LT : MP_GT;
     } else {
         cmp_b = bb->u.smallint.value < 0 ? MP_LT : MP_GT;
     }
@@ -562,9 +601,9 @@ MVMObject *MVM_bigint_div(MVMThreadContext *tc, MVMObject *result_type, MVMObjec
         ic = MVM_malloc(sizeof(mp_int));
         mp_init(ic);
 
-        // if we do a div with a negative, we need to make sure
-        // the result is floored rather than rounded towards
-        // zero, like C and libtommath would do.
+        /* if we do a div with a negative, we need to make sure
+         * the result is floored rather than rounded towards
+         * zero, like C and libtommath would do. */
         if ((cmp_a == MP_LT) ^ (cmp_b == MP_LT)) {
             mp_init(&remainder);
             mp_init(&intermediate);
@@ -591,6 +630,7 @@ MVMObject *MVM_bigint_div(MVMThreadContext *tc, MVMObject *result_type, MVMObjec
         }
         store_bigint_result(bc, ic);
         clear_temp_bigints(tmp, 2);
+        adjust_nursery(tc, bc);
     } else {
         MVMint32 num   = ba->u.smallint.value;
         MVMint32 denom = bb->u.smallint.value;
@@ -623,39 +663,46 @@ MVMObject * MVM_bigint_pow(MVMThreadContext *tc, MVMObject *a, MVMObject *b,
     mp_int *base        = force_bigint(ba, tmp);
     mp_int *exponent    = force_bigint(bb, tmp);
     mp_digit exponent_d = 0;
-    int cmp             = mp_cmp_d(exponent, 0);
 
-    if ((cmp == MP_EQ) || (MP_EQ == mp_cmp_d(base, 1))) {
+    if (mp_iszero(exponent) || (MP_EQ == mp_cmp_d(base, 1))) {
         r = MVM_repr_box_int(tc, int_type, 1);
     }
-    else if (cmp == MP_GT) {
-        mp_int *ic = MVM_malloc(sizeof(mp_int));
-        mp_init(ic);
+    else if (SIGN(exponent) == MP_ZPOS) {
         exponent_d = mp_get_int(exponent);
         if ((MP_GT == mp_cmp_d(exponent, exponent_d))) {
-            cmp = mp_cmp_d(base, 0);
-            if ((MP_EQ == cmp) || (MP_EQ == mp_cmp_d(base, 1))) {
-                mp_copy(base, ic);
+            if (mp_iszero(base)) {
+                r = MVM_repr_box_int(tc, int_type, 0);
+            }
+            else if (mp_get_int(base) == 1) {
+                r = MVM_repr_box_int(tc, int_type, MP_ZPOS == SIGN(base) || mp_iseven(exponent) ? 1 : -1);
             }
             else {
-                MVMnum64 ZERO = 0.0;
-                if (MP_GT == cmp) {
-                    mp_set_int(ic, (MVMnum64)1.0 / ZERO);
+                MVMnum64 inf;
+                if (MP_ZPOS == SIGN(base) || mp_iseven(exponent)) {
+                    inf = MVM_num_posinf(tc);
                 }
                 else {
-                    mp_set_int(ic, (MVMnum64)(-1.0) / ZERO);
+                    inf = MVM_num_neginf(tc);
                 }
+                r = MVM_repr_box_num(tc, num_type, inf);
             }
         }
         else {
+            mp_int *ic = MVM_malloc(sizeof(mp_int));
+            MVMP6bigintBody *resbody;
+            mp_init(ic);
+            MVM_gc_mark_thread_blocked(tc);
             mp_expt_d(base, exponent_d, ic);
+            MVM_gc_mark_thread_unblocked(tc);
+            r = MVM_repr_alloc_init(tc, int_type);
+            resbody = get_bigint_body(tc, r);
+            store_bigint_result(resbody, ic);
+            adjust_nursery(tc, resbody);
         }
-        r = MVM_repr_alloc_init(tc, int_type);
-        store_bigint_result(get_bigint_body(tc, r), ic);
     }
     else {
-        MVMnum64 f_base = mp_get_double(base);
-        MVMnum64 f_exp = mp_get_double(exponent);
+        MVMnum64 f_base = mp_get_double(base, 0);
+        MVMnum64 f_exp = mp_get_double(exponent, 0);
         r = MVM_repr_box_num(tc, num_type, pow(f_base, f_exp));
     }
     clear_temp_bigints(tmp, 2);
@@ -681,6 +728,7 @@ MVMObject *MVM_bigint_shl(MVMThreadContext *tc, MVMObject *result_type, MVMObjec
         two_complement_shl(ib, ia, n);
         store_bigint_result(bb, ib);
         clear_temp_bigints(tmp, 1);
+        adjust_nursery(tc, bb);
     } else {
         MVMint64 value;
         if (n < 0)
@@ -692,7 +740,17 @@ MVMObject *MVM_bigint_shl(MVMThreadContext *tc, MVMObject *result_type, MVMObjec
 
     return result;
 }
-
+/* Checks if a MVMP6bigintBody is negative. Handles cases where it is stored as
+ * a small int as well as cases when it is stored as a bigint */
+int BIGINT_IS_NEGATIVE (MVMP6bigintBody *ba) {
+    mp_int *mp_a = ba->u.bigint;
+    if (MVM_BIGINT_IS_BIG(ba)) {
+        return SIGN(mp_a) == MP_NEG;
+    }
+    else {
+        return ba->u.smallint.value < 0;
+    }
+}
 MVMObject *MVM_bigint_shr(MVMThreadContext *tc, MVMObject *result_type, MVMObject *a, MVMint64 n) {
     MVMP6bigintBody *ba = get_bigint_body(tc, a);
     MVMP6bigintBody *bb;
@@ -712,8 +770,9 @@ MVMObject *MVM_bigint_shr(MVMThreadContext *tc, MVMObject *result_type, MVMObjec
         two_complement_shl(ib, ia, -n);
         store_bigint_result(bb, ib);
         clear_temp_bigints(tmp, 1);
+        adjust_nursery(tc, bb);
     } else if (n >= 32) {
-        store_int64_result(bb, 0);
+        store_int64_result(bb, BIGINT_IS_NEGATIVE(ba) ? -1 : 0);
     } else {
         MVMint32 value = ba->u.smallint.value;
         value = value >> n;
@@ -742,6 +801,7 @@ MVMObject *MVM_bigint_not(MVMThreadContext *tc, MVMObject *result_type, MVMObjec
         mp_add_d(ia, 1, ib);
         mp_neg(ib, ib);
         store_bigint_result(bb, ib);
+        adjust_nursery(tc, bb);
     } else {
         MVMint32 value = ba->u.smallint.value;
         value = ~value;
@@ -768,23 +828,99 @@ void MVM_bigint_expmod(MVMThreadContext *tc, MVMObject *result, MVMObject *a, MV
     mp_exptmod(ia, ib, ic, id);
     store_bigint_result(bd, id);
     clear_temp_bigints(tmp, 3);
+    adjust_nursery(tc, bd);
 }
 
 void MVM_bigint_from_str(MVMThreadContext *tc, MVMObject *a, const char *buf) {
     MVMP6bigintBody *body = get_bigint_body(tc, a);
-    mp_int *i = MVM_malloc(sizeof(mp_int));
+    mp_int *i = alloca(sizeof(mp_int));
     mp_init(i);
     mp_read_radix(i, buf, 10);
+    adjust_nursery(tc, body);
     if (can_be_smallint(i)) {
         body->u.smallint.flag = MVM_BIGINT_32_FLAG;
-        body->u.smallint.value = SIGN(i) ? -DIGIT(i, 0) : DIGIT(i, 0);
+        body->u.smallint.value = SIGN(i) == MP_NEG ? -DIGIT(i, 0) : DIGIT(i, 0);
         mp_clear(i);
-        MVM_free(i);
     }
     else {
-        body->u.bigint = i;
+        mp_int *i_cpy = MVM_malloc(sizeof(mp_int));
+        memcpy(i_cpy, i, sizeof(mp_int));
+        body->u.bigint = i_cpy;
     }
 }
+#define can_fit_into_8bit(g) ((-128 <= (g) && (g) <= 127))
+MVMObject * MVM_coerce_sI(MVMThreadContext *tc, MVMString *s, MVMObject *type) {
+    char *buf = NULL;
+    int is_malloced = 0;
+    MVMStringIndex i;
+    MVMObject *a = MVM_repr_alloc_init(tc, type);
+    if (s->body.num_graphs < 120) {
+        buf = alloca(s->body.num_graphs + 1);
+    }
+    else {
+        buf = MVM_malloc(s->body.num_graphs + 1);
+        is_malloced = 1;
+    }
+    /* We just ignore synthetics since parsing will fail if a synthetic is
+     * encountered anyway. */
+    switch (s->body.storage_type) {
+        case MVM_STRING_GRAPHEME_ASCII:
+        case MVM_STRING_GRAPHEME_8:
+            memcpy(buf, s->body.storage.blob_8, sizeof(MVMGrapheme8) * s->body.num_graphs);
+            break;
+        case MVM_STRING_GRAPHEME_32:
+            for (i = 0; i < s->body.num_graphs; i++) {
+                buf[i] = can_fit_into_8bit(s->body.storage.blob_32[i])
+                    ? s->body.storage.blob_32[i]
+                    : '?'; /* Add a filler bogus char if it can't fit */
+            }
+            break;
+        case MVM_STRING_STRAND: {
+            MVMGraphemeIter gi;
+            MVM_string_gi_init(tc, &gi, s);
+            for (i = 0; i < s->body.num_graphs; i++) {
+                MVMGrapheme32 g = MVM_string_gi_get_grapheme(tc, &gi);
+                buf[i] = can_fit_into_8bit(g) ? g : '?';
+            }
+            break;
+        }
+        default:
+            if (is_malloced) MVM_free(buf);
+            MVM_exception_throw_adhoc(tc, "String corruption found in MVM_coerce_sI");
+    }
+    buf[s->body.num_graphs] = 0;
+
+    MVM_bigint_from_str(tc, a, buf);
+    if (is_malloced) MVM_free(buf);
+    return a;
+}
+
+MVMObject * MVM_bigint_from_bigint(MVMThreadContext *tc, MVMObject *result_type, MVMObject *a) {
+    MVMP6bigintBody *a_body;
+    MVMP6bigintBody *r_body;
+    MVMObject       *result;
+
+    MVMROOT(tc, a, {
+        result = MVM_repr_alloc_init(tc, result_type);
+    });
+
+    a_body = get_bigint_body(tc, a);
+    r_body = get_bigint_body(tc, result);
+
+    if (MVM_BIGINT_IS_BIG(a_body)) {
+        mp_int *i = MVM_malloc(sizeof(mp_int));
+        mp_init_copy(i, a_body->u.bigint);
+        store_bigint_result(r_body, i);
+        adjust_nursery(tc, r_body);
+    }
+    else {
+        r_body->u.smallint       = a_body->u.smallint;
+        r_body->u.smallint.flag  = a_body->u.smallint.flag;
+        r_body->u.smallint.value = a_body->u.smallint.value;
+    }
+
+    return result;
+}
 
 MVMString * MVM_bigint_to_str(MVMThreadContext *tc, MVMObject *a, int base) {
     MVMP6bigintBody *body = get_bigint_body(tc, a);
@@ -793,11 +929,18 @@ MVMString * MVM_bigint_to_str(MVMThreadContext *tc, MVMObject *a, int base) {
         int len;
         char *buf;
         MVMString *result;
+        int is_malloced = 0;
         mp_radix_size(i, base, &len);
-        buf = (char *) MVM_malloc(len);
+        if (len < 120) {
+            buf = alloca(len);
+        }
+        else {
+            buf = MVM_malloc(len);
+            is_malloced = 1;
+        }
         mp_toradix_n(i, buf, base, len);
         result = MVM_string_ascii_decode(tc, tc->instance->VMString, buf, len - 1);
-        MVM_free(buf);
+        if (is_malloced) MVM_free(buf);
         return result;
     }
     else {
@@ -808,11 +951,11 @@ MVMString * MVM_bigint_to_str(MVMThreadContext *tc, MVMObject *a, int base) {
             /* It's small, but shove it through bigint lib, as it knows how to
              * get other bases right. */
             mp_int i;
-            int len;
+            int len, is_malloced = 0;
             char *buf;
             MVMString *result;
 
-            MVMint32 value = body->u.smallint.value;
+            MVMint64 value = body->u.smallint.value;
             mp_init(&i);
             if (value >= 0) {
                 mp_set_long(&i, value);
@@ -823,10 +966,16 @@ MVMString * MVM_bigint_to_str(MVMThreadContext *tc, MVMObject *a, int base) {
             }
 
             mp_radix_size(&i, base, &len);
-            buf = (char *) MVM_malloc(len);
+            if (len < 120) {
+                buf = alloca(len);
+            }
+            else {
+                buf = MVM_malloc(len);
+                is_malloced = 1;
+            }
             mp_toradix_n(&i, buf, base, len);
             result = MVM_string_ascii_decode(tc, tc->instance->VMString, buf, len - 1);
-            MVM_free(buf);
+            if (is_malloced) MVM_free(buf);
             mp_clear(&i);
 
             return result;
@@ -839,7 +988,7 @@ MVMnum64 MVM_bigint_to_num(MVMThreadContext *tc, MVMObject *a) {
 
     if (MVM_BIGINT_IS_BIG(ba)) {
         mp_int *ia = ba->u.bigint;
-        return mp_get_double(ia);
+        return mp_get_double(ia, 0);
     } else {
         return (double)ba->u.smallint.value;
     }
@@ -865,18 +1014,32 @@ MVMnum64 MVM_bigint_div_num(MVMThreadContext *tc, MVMObject *a, MVMObject *b) {
         mp_int *ia = force_bigint(ba, tmp);
         mp_int *ib = force_bigint(bb, tmp);
 
-        int max_size = DIGIT_BIT * MAX(USED(ia), USED(ib));
-        if (max_size > 1023) {
-            mp_int reduced_a, reduced_b;
-            mp_init(&reduced_a);
-            mp_init(&reduced_b);
-            mp_div_2d(ia, max_size - 1023, &reduced_a, NULL);
-            mp_div_2d(ib, max_size - 1023, &reduced_b, NULL);
-            c = mp_get_double(&reduced_a) / mp_get_double(&reduced_b);
-            mp_clear(&reduced_a);
-            mp_clear(&reduced_b);
-        } else {
-            c = mp_get_double(ia) / mp_get_double(ib);
+        mp_clamp(ib);
+        if (ib->used == 0) { /* zero-denominator special case */
+            if (ia->sign == MP_NEG)
+                c = MVM_NUM_NEGINF;
+            else
+                c =  MVM_NUM_POSINF;
+            /*
+             * we won't have NaN case here, since the branch requires at
+             * least one bigint to be big
+             */
+        }
+        else {
+            mp_int scaled;
+            int bbits = mp_count_bits(ib)+64;
+
+            if (mp_init(&scaled) != MP_OKAY)
+                MVM_exception_throw_adhoc(tc,
+                    "Failed to initialize bigint for scaled divident");
+            if (mp_mul_2d(ia, bbits, &scaled) != MP_OKAY)
+                MVM_exception_throw_adhoc(tc, "Failed to scale divident");
+            // simply re-use &scaled for result
+            if (mp_div(&scaled, ib, &scaled, NULL) != MP_OKAY)
+                MVM_exception_throw_adhoc(tc,
+                    "Failed to preform bigint division");
+            c = mp_get_double(&scaled, bbits);
+            mp_clear(&scaled);
         }
         clear_temp_bigints(tmp, 2);
     } else {
@@ -885,19 +1048,65 @@ MVMnum64 MVM_bigint_div_num(MVMThreadContext *tc, MVMObject *a, MVMObject *b) {
     return c;
 }
 
-void MVM_bigint_rand(MVMThreadContext *tc, MVMObject *a, MVMObject *b) {
-    MVMP6bigintBody *ba = get_bigint_body(tc, a);
+MVMObject * MVM_bigint_rand(MVMThreadContext *tc, MVMObject *type, MVMObject *b) {
+    MVMObject *result;
+    MVMP6bigintBody *ba;
     MVMP6bigintBody *bb = get_bigint_body(tc, b);
 
-    mp_int *tmp[1] = { NULL };
-    mp_int *rnd = MVM_malloc(sizeof(mp_int));
-    mp_int *max = force_bigint(bb, tmp);
+    MVMint8 use_small_arithmetic = 0;
+    MVMint8 have_to_negate = 0;
+    MVMint32 smallint_max = 0;
 
-    mp_init(rnd);
-    mp_rand(rnd, USED(max) + 1);
-    mp_mod(rnd, max, rnd);
-    store_bigint_result(ba, rnd);
-    clear_temp_bigints(tmp, 1);
+    if (MVM_BIGINT_IS_BIG(bb)) {
+        if (can_be_smallint(bb->u.bigint)) {
+            use_small_arithmetic = 1;
+            smallint_max = DIGIT(bb->u.bigint, 0);
+            have_to_negate = SIGN(bb->u.bigint) == MP_NEG;
+        }
+    } else {
+        use_small_arithmetic = 1;
+        smallint_max = bb->u.smallint.value;
+    }
+
+    if (use_small_arithmetic) {
+        if (MP_GEN_RANDOM_MAX >= abs(smallint_max)) {
+            mp_digit result_int = MP_GEN_RANDOM();
+            result_int = result_int % smallint_max;
+            if(have_to_negate)
+                result_int *= -1;
+
+            MVMROOT2(tc, type, b, {
+                result = MVM_repr_alloc_init(tc, type);
+            });
+
+            ba = get_bigint_body(tc, result);
+            store_int64_result(ba, result_int);
+        } else {
+            use_small_arithmetic = 0;
+        }
+    }
+
+    if (!use_small_arithmetic) {
+        mp_int *tmp[1] = { NULL };
+        mp_int *rnd = MVM_malloc(sizeof(mp_int));
+        mp_int *max = force_bigint(bb, tmp);
+
+        MVMROOT2(tc, type, b, {
+            result = MVM_repr_alloc_init(tc, type);
+        });
+
+        ba = get_bigint_body(tc, result);
+
+        mp_init(rnd);
+        mp_rand(rnd, USED(max) + 1);
+
+        mp_mod(rnd, max, rnd);
+        store_bigint_result(ba, rnd);
+        clear_temp_bigints(tmp, 1);
+        adjust_nursery(tc, ba);
+    }
+
+    return result;
 }
 
 MVMint64 MVM_bigint_is_prime(MVMThreadContext *tc, MVMObject *a, MVMint64 b) {
@@ -920,8 +1129,8 @@ MVMint64 MVM_bigint_is_prime(MVMThreadContext *tc, MVMObject *a, MVMint64 b) {
             return result;
         }
     } else {
-        // we only reach this if we have a smallint that's equal to 1.
-        // which we define as not-prime.
+        /* we only reach this if we have a smallint that's equal to 1.
+         * which we define as not-prime. */
         return 0;
     }
 }
@@ -995,18 +1204,17 @@ MVMObject * MVM_bigint_radix(MVMThreadContext *tc, MVMint64 radix, MVMString *st
         else if (ch >= 'A' && ch <= 'Z') ch = ch - 'A' + 10;
         else if (ch >= 0xFF21 && ch <= 0xFF3A) ch = ch - 0xFF21 + 10; /* uppercase fullwidth */
         else if (ch >= 0xFF41 && ch <= 0xFF5A) ch = ch - 0xFF41 + 10; /* lowercase fullwidth */
-        else if (ch > 0 && MVM_unicode_codepoint_has_property_value(tc, ch, MVM_UNICODE_PROPERTY_GENERAL_CATEGORY,
-                MVM_unicode_cname_to_property_value_code(tc, MVM_UNICODE_PROPERTY_GENERAL_CATEGORY, STR_WITH_LEN("Nd")))) {
-            /* as of Unicode 6.0.0, characters with the 'de' Numeric Type (and are
+        else if (ch > 0 && MVM_unicode_codepoint_get_property_int(tc, ch, MVM_UNICODE_PROPERTY_NUMERIC_TYPE)
+         == MVM_UNICODE_PVALUE_Numeric_Type_DECIMAL) {
+            /* as of Unicode 9.0.0, characters with the 'de' Numeric Type (and are
              * thus also of General Category Nd, since 4.0.0) are contiguous
              * sequences of 10 chars whose Numeric Values ascend from 0 through 9.
              */
 
-            /* the string returned for NUMERIC_VALUE contains a floating point
-             * value, so atoi will stop on the . in the string. This is fine
-             * though, since we'd have to truncate the float regardless.
-             */
-            ch = atoi(MVM_unicode_codepoint_get_property_cstr(tc, ch, MVM_UNICODE_PROPERTY_NUMERIC_VALUE));
+            /* the string returned for NUMERIC_VALUE_NUMERATOR contains an integer
+             * value. We can use numerator because they all are from 0-9 and have
+             * denominator of 1 */
+            ch = fast_atoi(MVM_unicode_codepoint_get_property_cstr(tc, ch, MVM_UNICODE_PROPERTY_NUMERIC_VALUE_NUMERATOR));
         }
         else break;
         if (ch >= radix) break;
@@ -1033,6 +1241,9 @@ MVMObject * MVM_bigint_radix(MVMThreadContext *tc, MVMint64 radix, MVMString *st
     store_bigint_result(bvalue, value);
     store_bigint_result(bbase, base);
 
+    adjust_nursery(tc, bvalue);
+    adjust_nursery(tc, bbase);
+
     pos_obj = MVM_repr_box_int(tc, type, pos);
     MVM_repr_push_o(tc, result, pos_obj);
 
@@ -1054,7 +1265,7 @@ MVMint64 MVM_bigint_is_big(MVMThreadContext *tc, MVMObject *a) {
             is_big = 1;
         return is_big;
     } else {
-        // if it's in a smallint, it's 32 bits big at most and fits into an INTVAL easily.
+        /* if it's in a smallint, it's 32 bits big at most and fits into an INTVAL easily. */
         return 0;
     }
 }
