@@ -81,21 +81,21 @@ MVMInstance * MVM_vm_create_instance(void) {
     MVMInstance *instance;
 
     char *spesh_log, *spesh_nodelay, *spesh_disable, *spesh_inline_disable,
-         *spesh_osr_disable, *spesh_limit, *spesh_blocking;
+         *spesh_osr_disable, *spesh_limit, *spesh_blocking, *spesh_inline_log;
     char *jit_log, *jit_expr_disable, *jit_disable, *jit_bytecode_dir, *jit_last_frame, *jit_last_bb;
     char *dynvar_log;
     int init_stat;
-    MVMuint32 hashSecret;
-    MVMuint64 now = MVM_platform_now();
 
     /* Set up instance data structure. */
     instance = MVM_calloc(1, sizeof(MVMInstance));
 
     /* Create the main thread's ThreadContext and stash it. */
     instance->main_thread = MVM_tc_create(NULL, instance);
-    MVM_getrandom(instance->main_thread, &hashSecret, sizeof(MVMuint32));
-    instance->hashSecret ^= now;
-    instance->hashSecret ^= MVM_proc_getpid(instance->main_thread) * now;
+    /* Get the 128-bit hashSecret */
+    MVM_getrandom(instance->main_thread, instance->hashSecrets, sizeof(MVMuint64) * 2);
+    /* Just in case MVM_getrandom didn't work, XOR it with some (poorly) randomized data */
+    instance->hashSecrets[0] ^= ptr_hash_64_to_64((MVMuint64)instance);
+    instance->hashSecrets[1] ^= MVM_proc_getpid(instance->main_thread) * MVM_platform_now();
     instance->main_thread->thread_id = 1;
 
     /* Next thread to be created gets ID 2 (the main thread got ID 1). */
@@ -115,6 +115,9 @@ MVMInstance * MVM_vm_create_instance(void) {
     init_cond(instance->cond_gc_intrays_clearing, "GC intrays clearing");
     init_cond(instance->cond_blocked_can_continue, "GC thread unblock");
 
+    /* Safe point free list. */
+    init_mutex(instance->mutex_free_at_safepoint, "safepoint free list");
+
     /* Create fixed size allocator. */
     instance->fsa = MVM_fixed_size_create(instance->main_thread);
 
@@ -245,6 +248,11 @@ MVMInstance * MVM_vm_create_instance(void) {
     if (spesh_blocking && spesh_blocking[0])
         instance->spesh_blocking = 1;
 
+    /* Should we dump details of inlining? */
+    spesh_inline_log = getenv("MVM_SPESH_INLINE_LOG");
+    if (spesh_inline_log && spesh_inline_log[0])
+        instance->spesh_inline_log = 1;
+
     /* JIT environment/logging setup. */
     jit_disable = getenv("MVM_JIT_DISABLE");
     if (!jit_disable || !jit_disable[0])
@@ -273,7 +281,7 @@ MVMInstance * MVM_vm_create_instance(void) {
     /* what could possibly go wrong in integer formats? */
     instance->jit_expr_last_frame = jit_last_frame != NULL ? atoi(jit_last_frame) : -1;
     instance->jit_expr_last_bb    =    jit_last_bb != NULL ? atoi(jit_last_bb) : -1;
-    instance->jit_seq_nr = 0;
+    instance->jit_seq_nr = 1;
 
     /* add JIT debugging breakpoints */
     {
@@ -502,7 +510,7 @@ void MVM_vm_destroy_instance(MVMInstance *instance) {
 
     /* Cleanup REPR registry */
     uv_mutex_destroy(&instance->mutex_repr_registry);
-    MVM_HASH_DESTROY(hash_handle, MVMReprRegistry, instance->repr_hash);
+    MVM_HASH_DESTROY(instance->main_thread, hash_handle, MVMReprRegistry, instance->repr_hash);
     MVM_free(instance->repr_list);
 
     /* Clean up GC related resources. */
@@ -515,35 +523,39 @@ void MVM_vm_destroy_instance(MVMInstance *instance) {
     uv_cond_destroy(&instance->cond_blocked_can_continue);
     uv_mutex_destroy(&instance->mutex_gc_orchestrate);
 
+    /* Clean up safepoint free vector. */
+    MVM_VECTOR_DESTROY(instance->free_at_safepoint);
+    uv_mutex_destroy(&instance->mutex_free_at_safepoint);
+
     /* Clean up Hash of HLLConfig. */
     uv_mutex_destroy(&instance->mutex_hllconfigs);
-    MVM_HASH_DESTROY(hash_handle, MVMHLLConfig, instance->compiler_hll_configs);
-    MVM_HASH_DESTROY(hash_handle, MVMHLLConfig, instance->compilee_hll_configs);
+    MVM_HASH_DESTROY(instance->main_thread, hash_handle, MVMHLLConfig, instance->compiler_hll_configs);
+    MVM_HASH_DESTROY(instance->main_thread, hash_handle, MVMHLLConfig, instance->compilee_hll_configs);
 
     /* Clean up Hash of DLLs. */
     uv_mutex_destroy(&instance->mutex_dll_registry);
-    MVM_HASH_DESTROY(hash_handle, MVMDLLRegistry, instance->dll_registry);
+    MVM_HASH_DESTROY(instance->main_thread, hash_handle, MVMDLLRegistry, instance->dll_registry);
 
     /* Clean up Hash of extensions. */
     uv_mutex_destroy(&instance->mutex_ext_registry);
-    MVM_HASH_DESTROY(hash_handle, MVMExtRegistry, instance->ext_registry);
+    MVM_HASH_DESTROY(instance->main_thread, hash_handle, MVMExtRegistry, instance->ext_registry);
 
     /* Clean up Hash of extension ops. */
     uv_mutex_destroy(&instance->mutex_extop_registry);
-    MVM_HASH_DESTROY(hash_handle, MVMExtOpRegistry, instance->extop_registry);
+    MVM_HASH_DESTROY(instance->main_thread, hash_handle, MVMExtOpRegistry, instance->extop_registry);
 
     /* Clean up Hash of all known serialization contexts; all SCs list is in
      * FSA space and so cleaned up with that. */
     uv_mutex_destroy(&instance->mutex_sc_registry);
-    MVM_HASH_DESTROY(hash_handle, MVMSerializationContextBody, instance->sc_weakhash);
+    MVM_HASH_DESTROY(instance->main_thread, hash_handle, MVMSerializationContextBody, instance->sc_weakhash);
 
     /* Clean up Hash of filenames of compunits loaded from disk. */
     uv_mutex_destroy(&instance->mutex_loaded_compunits);
-    MVM_HASH_DESTROY(hash_handle, MVMLoadedCompUnitName, instance->loaded_compunits);
+    MVM_HASH_DESTROY(instance->main_thread, hash_handle, MVMLoadedCompUnitName, instance->loaded_compunits);
 
     /* Clean up Container registry. */
     uv_mutex_destroy(&instance->mutex_container_registry);
-    MVM_HASH_DESTROY(hash_handle, MVMContainerRegistry, instance->container_registry);
+    MVM_HASH_DESTROY(instance->main_thread, hash_handle, MVMContainerRegistry, instance->container_registry);
 
     /* Clean up Hash of compiler objects keyed by name. */
     uv_mutex_destroy(&instance->mutex_compiler_registry);
