@@ -13,11 +13,6 @@
 /* platform-specific setjmp override */
 #include <platform/setjmp.h>
 
-/* stuff for uthash */
-#define uthash_fatal(msg) MVM_exception_throw_adhoc(tc, "internal hash error: " msg)
-
-#include <uthash.h>
-
 /* libuv
  * must precede atomic_ops.h so we get the ordering of Winapi headers right
  */
@@ -87,6 +82,11 @@ typedef double   MVMnum64;
 #  define MVM_USED_BY_JIT
 #endif
 
+/* stuff for uthash */
+#define uthash_fatal(msg) MVM_exception_throw_adhoc(tc, "internal hash error: " msg)
+
+#include "strings/uthash_types.h"
+
 MVM_PUBLIC const MVMint32 MVM_jit_support(void);
 
 /* Headers for various other data structures and APIs. */
@@ -97,6 +97,7 @@ MVM_PUBLIC const MVMint32 MVM_jit_support(void);
 #include "core/vector.h"
 #include "core/threadcontext.h"
 #include "core/instance.h"
+#include "strings/uthash.h"
 #include "core/interp.h"
 #include "core/callsite.h"
 #include "core/args.h"
@@ -141,14 +142,17 @@ MVM_PUBLIC const MVMint32 MVM_jit_support(void);
 #include "gc/finalize.h"
 #include "core/regionalloc.h"
 #include "spesh/dump.h"
+#include "spesh/debug.h"
 #include "spesh/graph.h"
 #include "spesh/codegen.h"
 #include "spesh/candidate.h"
 #include "spesh/manipulate.h"
 #include "spesh/args.h"
+#include "spesh/usages.h"
 #include "spesh/facts.h"
 #include "spesh/optimize.h"
 #include "spesh/dead_bb_elimination.h"
+#include "spesh/dead_ins_elimination.h"
 #include "spesh/deopt.h"
 #include "spesh/log.h"
 #include "spesh/threshold.h"
@@ -160,6 +164,8 @@ MVM_PUBLIC const MVMint32 MVM_jit_support(void);
 #include "spesh/stats.h"
 #include "spesh/plan.h"
 #include "spesh/arg_guard.h"
+#include "spesh/plugin.h"
+#include "spesh/frame_walker.h"
 #include "strings/nfg.h"
 #include "strings/normalize.h"
 #include "strings/decode_stream.h"
@@ -175,6 +181,7 @@ MVM_PUBLIC const MVMint32 MVM_jit_support(void);
 #include "strings/latin1.h"
 #include "strings/windows1252.h"
 #include "strings/shiftjis.h"
+#include "strings/unicode_ops.h"
 #include "io/io.h"
 #include "io/eventloop.h"
 #include "io/syncfile.h"
@@ -198,6 +205,7 @@ MVM_PUBLIC const MVMint32 MVM_jit_support(void);
 #include "jit/tile.h"
 #include "jit/compile.h"
 #include "jit/log.h"
+#include "jit/interface.h"
 #include "profiler/instrument.h"
 #include "profiler/log.h"
 #include "profiler/profile.h"
@@ -212,7 +220,7 @@ MVMObject *MVM_backend_config(MVMThreadContext *tc);
 MVM_PUBLIC MVMInstance * MVM_vm_create_instance(void);
 MVM_PUBLIC void MVM_vm_run_file(MVMInstance *instance, const char *filename);
 MVM_PUBLIC void MVM_vm_dump_file(MVMInstance *instance, const char *filename);
-MVM_PUBLIC MVM_NO_RETURN void MVM_vm_exit(MVMInstance *instance) MVM_NO_RETURN_GCC;
+MVM_PUBLIC MVM_NO_RETURN void MVM_vm_exit(MVMInstance *instance) MVM_NO_RETURN_ATTRIBUTE;
 MVM_PUBLIC void MVM_vm_destroy_instance(MVMInstance *instance);
 MVM_PUBLIC void MVM_vm_set_clargs(MVMInstance *instance, int argc, char **argv);
 MVM_PUBLIC void MVM_vm_set_exec_name(MVMInstance *instance, const char *exec_name);
