@@ -59,8 +59,34 @@ typedef struct {
     MVMuint32 inlined_size;
 } SpeshGraphSizeStats;
 
+typedef struct {
+    MVMint32 cur_depth;
+    MVMint32 inline_idx[64];
+} InlineIndexStack;
+
+static void push_inline(MVMThreadContext *tc, InlineIndexStack *stack, MVMint32 idx) {
+    if (stack->cur_depth == 63)
+        MVM_oops(tc, "Too many levels of inlining to dump");
+    stack->cur_depth++;
+    stack->inline_idx[stack->cur_depth] = idx;
+}
+
+static void pop_inline(MVMThreadContext *tc, InlineIndexStack *stack) {
+    stack->cur_depth--;
+    if (stack->cur_depth < -1)
+        MVM_oops(tc, "Too many levels of inlining popped");
+}
+
+static MVMCompUnit * get_current_cu(MVMThreadContext *tc, MVMSpeshGraph *g, InlineIndexStack *stack) {
+    if (stack->cur_depth < 0)
+        return g->sf->body.cu;
+    else
+        return g->inlines[stack->inline_idx[stack->cur_depth]].sf->body.cu;
+}
+
 /* Dumps a basic block. */
-static void dump_bb(MVMThreadContext *tc, DumpStr *ds, MVMSpeshGraph *g, MVMSpeshBB *bb, SpeshGraphSizeStats *stats) {
+static void dump_bb(MVMThreadContext *tc, DumpStr *ds, MVMSpeshGraph *g, MVMSpeshBB *bb,
+                    SpeshGraphSizeStats *stats, InlineIndexStack *inline_stack) {
     MVMSpeshIns *cur_ins;
     MVMint64     i;
     MVMint32     size = 0;
@@ -131,10 +157,12 @@ static void dump_bb(MVMThreadContext *tc, DumpStr *ds, MVMSpeshGraph *g, MVMSpes
                 case MVM_SPESH_ANN_INLINE_START:
                     appendf(ds, "      [Annotation: Inline Start (%d)]\n",
                         ann->data.inline_idx);
+                    push_inline(tc, inline_stack, ann->data.inline_idx);
                     break;
                 case MVM_SPESH_ANN_INLINE_END:
                     appendf(ds, "      [Annotation: Inline End (%d)]\n",
                         ann->data.inline_idx);
+                    pop_inline(tc, inline_stack);
                     break;
                 case MVM_SPESH_ANN_DEOPT_INLINE:
                     appendf(ds, "      [Annotation: INS Deopt Inline (idx %d -> pc %d; line %d)]\n",
@@ -146,7 +174,8 @@ static void dump_bb(MVMThreadContext *tc, DumpStr *ds, MVMSpeshGraph *g, MVMSpes
                     break;
                 case MVM_SPESH_ANN_LINENO: {
                     char *cstr = MVM_string_utf8_encode_C_string(tc,
-                        MVM_cu_string(tc, g->sf->body.cu, ann->data.lineno.filename_string_index));
+                        MVM_cu_string(tc, get_current_cu(tc, g, inline_stack),
+                        ann->data.lineno.filename_string_index));
                     appendf(ds, "      [Annotation: Line Number: %s:%d]\n",
                         cstr, ann->data.lineno.line_number);
                     MVM_free(cstr);
@@ -156,6 +185,10 @@ static void dump_bb(MVMThreadContext *tc, DumpStr *ds, MVMSpeshGraph *g, MVMSpes
                     appendf(ds, "      [Annotation: Logged (bytecode offset %d)]\n",
                         ann->data.bytecode_offset);
                     break;
+                case MVM_SPESH_ANN_DEOPT_SYNTH:
+                    appendf(ds, "      [Annotation: INS Deopt Synth (idx %d)]\n",
+                        ann->data.deopt_idx);
+                    break;
                 default:
                     appendf(ds, "      [Annotation: %d (unknown)]\n", ann->type);
             }
@@ -362,6 +395,23 @@ static void dump_bb(MVMThreadContext *tc, DumpStr *ds, MVMSpeshGraph *g, MVMSpes
     append(ds, "\n\n");
 }
 
+/* Dump deopt usages. */
+static void dump_deopt_usages(MVMThreadContext *tc, DumpStr *ds, MVMSpeshGraph *g, MVMSpeshOperand operand) {
+    MVMSpeshFacts *facts = MVM_spesh_get_facts(tc, g, operand);
+    MVMSpeshDeoptUseEntry *entry = facts->usage.deopt_users;
+    if (entry) {
+        MVMuint32 first = 1;
+        append(ds, ", deopt=");
+        while (entry) {
+            if (first)
+                first = 0;
+            else
+                append(ds, ",");
+            appendf(ds, "%d", entry->deopt_idx);
+            entry = entry->next;
+        }
+    }
+}
 /* Dumps the facts table. */
 static void dump_facts(MVMThreadContext *tc, DumpStr *ds, MVMSpeshGraph *g) {
     MVMuint16 i, j, num_locals, num_facts;
@@ -369,12 +419,18 @@ static void dump_facts(MVMThreadContext *tc, DumpStr *ds, MVMSpeshGraph *g) {
     for (i = 0; i < num_locals; i++) {
         num_facts = g->fact_counts[i];
         for (j = 0; j < num_facts; j++) {
-            MVMint32 usages = g->facts[i][j].usages;
-            MVMint32 flags  = g->facts[i][j].flags;
+            MVMint32 flags = g->facts[i][j].flags;
+            MVMSpeshOperand operand;
+            operand.reg.orig = i;
+            operand.reg.i = j;
             if (i < 10) append(ds, " ");
             if (j < 10) append(ds, " ");
-            if (flags || g->facts[i][j].dead_writer || g->facts[i][j].writer && g->facts[i][j].writer->info->opcode == MVM_SSA_PHI) {
-                appendf(ds, "    r%d(%d): usages=%d, flags=%-5d", i, j, usages, flags);
+            if (flags || g->facts[i][j].dead_writer || (g->facts[i][j].writer && g->facts[i][j].writer->info->opcode == MVM_SSA_PHI)) {
+                appendf(ds, "    r%d(%d): usages=%d%s", i, j,
+                    MVM_spesh_usages_count(tc, g, operand),
+                    MVM_spesh_usages_is_used_by_handler(tc, g, operand) ? "+handler" : "");
+                dump_deopt_usages(tc, ds, g, operand);
+                appendf(ds, ", flags=%-5d", flags);
                 if (flags & 1) {
                     append(ds, " KnTyp");
                 }
@@ -424,8 +480,13 @@ static void dump_facts(MVMThreadContext *tc, DumpStr *ds, MVMSpeshGraph *g) {
                     appendf(ds, " (merged from %d regs)", g->facts[i][j].writer->info->num_operands - 1);
                 }
             }
-            else
-                appendf(ds, "    r%d(%d): usages=%d, flags=%d", i, j, usages, flags);
+            else {
+                appendf(ds, "    r%d(%d): usages=%d%s", i, j,
+                    MVM_spesh_usages_count(tc, g, operand),
+                    MVM_spesh_usages_is_used_by_handler(tc, g, operand) ? "+handler" : "");
+                dump_deopt_usages(tc, ds, g, operand);
+                appendf(ds, ", flags=%-5d", flags);
+            }
             append(ds, "\n");
         }
         append(ds, "\n");
@@ -487,10 +548,12 @@ static void dump_fileinfo(MVMThreadContext *tc, DumpStr *ds, MVMStaticFrame *sf)
 char * MVM_spesh_dump(MVMThreadContext *tc, MVMSpeshGraph *g) {
     MVMSpeshBB *cur_bb;
     SpeshGraphSizeStats stats;
+    InlineIndexStack inline_stack;
     DumpStr ds;
 
     stats.total_size = 0;
     stats.inlined_size = 0;
+    inline_stack.cur_depth = -1;
 
     /* Allocate buffer. */
     ds.alloc  = 8192;
@@ -513,7 +576,7 @@ char * MVM_spesh_dump(MVMThreadContext *tc, MVMSpeshGraph *g) {
     /* Go over all the basic blocks and dump them. */
     cur_bb = g->entry;
     while (cur_bb) {
-        dump_bb(tc, &ds, g, cur_bb, &stats);
+        dump_bb(tc, &ds, g, cur_bb, &stats, &inline_stack);
         cur_bb = cur_bb->linear_next;
     }
 
@@ -537,9 +600,28 @@ char * MVM_spesh_dump(MVMThreadContext *tc, MVMSpeshGraph *g) {
             else if (value->flags & MVM_CF_TYPE_OBJECT)
                 appendf(&ds, "    %d = Type Object (%s)\n", i,
                     MVM_6model_get_debug_name(tc, (MVMObject *)value));
-            else
-                appendf(&ds, "    %d = Instance (%s)\n", i,
-                    MVM_6model_get_debug_name(tc, (MVMObject *)value));
+            else {
+                MVMObject *obj = (MVMObject *)value;
+                MVMuint32 repr_id = REPR(obj)->ID;
+                appendf(&ds, "    %d = Instance (%s)", i,
+                    MVM_6model_get_debug_name(tc, obj));
+                if (repr_id == MVM_REPR_ID_MVMStaticFrame || repr_id == MVM_REPR_ID_MVMCode) {
+                    MVMStaticFrameBody *body;
+                    char *name_str;
+                    char *cuuid_str;
+                    if (repr_id == MVM_REPR_ID_MVMCode) {
+                        MVMCodeBody *code_body = (MVMCodeBody *)OBJECT_BODY(obj);
+                        obj = (MVMObject *)code_body->sf;
+                    }
+                    body = (MVMStaticFrameBody *)OBJECT_BODY(obj);
+                    name_str  = MVM_string_utf8_encode_C_string(tc, body->name);
+                    cuuid_str = MVM_string_utf8_encode_C_string(tc, body->cuuid);
+                    appendf(&ds, " - '%s' (%s)", name_str, cuuid_str);
+                    MVM_free(name_str);
+                    MVM_free(cuuid_str);
+                }
+                appendf(&ds, "\n");
+            }
         }
     }
 
@@ -628,6 +710,10 @@ void dump_stats_by_callsite(MVMThreadContext *tc, DumpStr *ds, MVMSpeshStatsByCa
                         oss->type_tuples[k].arg_types,
                         "                    ");
                 }
+                for (k = 0; k < oss->num_plugin_guards; k++)
+                    appendf(ds, "                %d x spesh plugin guard index %d\n",
+                        oss->plugin_guards[k].count,
+                        oss->plugin_guards[k].guard_index);
             }
         }
         append(ds, "\n");
@@ -731,6 +817,8 @@ char * MVM_spesh_dump_planned(MVMThreadContext *tc, MVMSpeshPlanned *p) {
                     p->cs_stats->osr_hits);
             else
                 append(&ds, "It was planned for unknown reasons.\n");
+            if (!p->sf->body.specializable)
+                append(&ds, "The body contains no specializable instructions.\n");
             break;
         case MVM_SPESH_PLANNED_OBSERVED_TYPES: {
             MVMCallsite *cs = p->cs_stats->cs;
