@@ -43,127 +43,113 @@ static void demand_extop(MVMThreadContext *tc, MVMCompUnit *target_cu, MVMCompUn
     MVM_oops(tc, "Spesh: inline failed to find source CU extop entry");
 }
 
-/* Sees if it will be possible to inline the target code ref, given we could
- * already identify a spesh candidate. Returns NULL if no inlining is possible
- * or a graph ready to be merged if it will be possible. */
-MVMSpeshGraph * MVM_spesh_inline_try_get_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
-                                               MVMStaticFrame *target_sf,
-                                               MVMSpeshCandidate *cand,
-                                               MVMSpeshIns *invoke_ins,
-                                               char **no_inline_reason) {
-    MVMSpeshGraph *ig;
-    MVMSpeshBB *bb;
-    MVMint32 same_hll;
-
+/* Considers a static frame and decides if it's possible to inline it into the
+ * current inliner. If there are no blockers, non-zero. Otherwise, sets the
+ * reason for not inlining and returns zero. */
+static int is_static_frame_inlineable(MVMThreadContext *tc, MVMSpeshGraph *inliner,
+                                      MVMStaticFrame *target_sf, char **no_inline_reason) {
     /* Check inlining is enabled. */
     if (!tc->instance->spesh_inline_enabled) {
         *no_inline_reason = "inlining is disabled";
-        return NULL;
+        return 0;
     }
 
     /* Check frame is not marked as not being allowed to be inlined. */
     if (target_sf->body.no_inline) {
         *no_inline_reason = "the frame is marked as no-inline";
-        return NULL;
-    }
-
-    /* Check bytecode size is within the inline limit. */
-    if (cand->bytecode_size > MVM_SPESH_MAX_INLINE_SIZE) {
-        *no_inline_reason = "bytecode is too large to inline";
-        return NULL;
+        return 0;
     }
 
     /* Ensure that this isn't a recursive inlining. */
     if (target_sf == inliner->sf) {
         *no_inline_reason = "recursive calls cannot be inlined";
-        return NULL;
+        return 0;
     }
 
     /* Ensure it has no state vars (these need the setup code in frame
      * invoke). */
     if (target_sf->body.has_state_vars) {
         *no_inline_reason = "cannot inline code that declares a state variable";
-        return NULL;
+        return 0;
     }
 
     /* Ensure it's not a thunk (need to skip over those in exception search). */
     if (target_sf->body.is_thunk) {
         *no_inline_reason = "cannot inline code marked as a thunk";
-        return NULL;
+        return 0;
     }
 
-    /* If they're from the same HLL, we'll need to watch out for ops that are
-     * HLL sensitive. */
-    same_hll = target_sf->body.cu->body.hll_config == inliner->sf->body.cu->body.hll_config;
+    return 1;
+}
 
-    /* Build graph from the already-specialized bytecode. */
-    ig = MVM_spesh_graph_create_from_cand(tc, target_sf, cand, 0);
-
-    /* Traverse graph, looking for anything that might prevent inlining and
-     * also building usage counts up. */
-    bb = ig->entry;
+/* Checks a spesh graph to see if it contains any uninlineable elements. If
+ * it is possbile to inline, returns non-zero. Otherwise, sets no_inline_reason
+ * and returns zero. Optionally builds up usage counts on the graph. */
+static int is_graph_inlineable(MVMThreadContext *tc, MVMSpeshGraph *inliner,
+                               MVMStaticFrame *target_sf, MVMSpeshIns *invoke_ins,
+                               MVMSpeshGraph *ig, char **no_inline_reason) {
+    MVMSpeshBB *bb = ig->entry;
+    MVMint32 same_hll = target_sf->body.cu->body.hll_config ==
+            inliner->sf->body.cu->body.hll_config;
     while (bb) {
         MVMSpeshIns *ins = bb->first_ins;
         while (ins) {
-            /* Track usages. */
             MVMint32 opcode = ins->info->opcode;
             MVMint32 is_phi = opcode == MVM_SSA_PHI;
-            MVMuint8 i;
-            for (i = 0; i < ins->info->num_operands; i++)
-                if ((is_phi && i > 0)
-                    || (!is_phi && (ins->info->operands[i] & MVM_operand_rw_mask) == MVM_operand_read_reg))
-                    ig->facts[ins->operands[i].reg.orig][ins->operands[i].reg.i].usages++;
-            if (opcode == MVM_OP_inc_i || opcode == MVM_OP_inc_u ||
-                    opcode == MVM_OP_dec_i || opcode == MVM_OP_dec_u)
-                ig->facts[ins->operands[0].reg.orig][ins->operands[0].reg.i - 1].usages++;
 
             /* Instruction may be marked directly as not being inlinable, in
              * which case we're done. */
             if (!is_phi && ins->info->no_inline) {
                 *no_inline_reason = "target has a :noinline instruction";
-                goto not_inlinable;
+                return 0;
+            }
+
+            /* XXX TODO: See if the workaround below is still required */
+            if (opcode == MVM_OP_throwpayloadlexcaller && tc->instance->profiling) {
+                *no_inline_reason = "target has throwpayloadlexcaller, which currently causes problems when profiling is on";
+                return 0;
             }
 
             /* If we don't have the same HLL and there's a :useshll op, we
              * cannot inline. */
             if (!same_hll && ins->info->uses_hll) {
                 *no_inline_reason = "target has a :useshll instruction and HLLs are different";
-                goto not_inlinable;
+                return 0;
             }
 
             /* If we have an invoke_o, but a return_[ins] that would require
              * boxing, we can't inline if it's not the same HLL. */
             if (!same_hll && invoke_ins->info->opcode == MVM_OP_invoke_o) {
-                switch (ins->info->opcode) {
+                switch (opcode) {
                     case MVM_OP_return_i:
                     case MVM_OP_return_n:
                     case MVM_OP_return_s:
                         *no_inline_reason = "target needs a return boxing and HLLs are different";
-                        goto not_inlinable;
+                        return 0;
                 }
             }
 
             /* If we have lexical bind, make sure it's within the frame. */
-            if (ins->info->opcode == MVM_OP_bindlex) {
+            if (opcode == MVM_OP_bindlex) {
                 if (ins->operands[0].lex.outers > 0) {
                     *no_inline_reason = "target has bind to outer lexical";
-                    goto not_inlinable;
+                    return 0;
                 }
             }
 
             /* Check we don't have too many args for inlining to work out. */
-            else if (ins->info->opcode == MVM_OP_sp_getarg_o ||
-                    ins->info->opcode == MVM_OP_sp_getarg_i ||
-                    ins->info->opcode == MVM_OP_sp_getarg_n ||
-                    ins->info->opcode == MVM_OP_sp_getarg_s) {
+            else if (opcode == MVM_OP_sp_getarg_o ||
+                    opcode == MVM_OP_sp_getarg_i ||
+                    opcode == MVM_OP_sp_getarg_n ||
+                    opcode == MVM_OP_sp_getarg_s) {
                 if (ins->operands[1].lit_i16 >= MAX_ARGS_FOR_OPT) {
                     *no_inline_reason = "too many arguments to inline";
-                    goto not_inlinable;
+                    return 0;
                 }
             }
 
             /* Ext-ops need special care in inter-comp-unit inlines. */
-            if (ins->info->opcode == (MVMuint16)-1) {
+            if (opcode == (MVMuint16)-1) {
                 MVMCompUnit *target_cu = inliner->sf->body.cu;
                 MVMCompUnit *source_cu = target_sf->body.cu;
                 if (source_cu != target_cu)
@@ -175,14 +161,92 @@ MVMSpeshGraph * MVM_spesh_inline_try_get_graph(MVMThreadContext *tc, MVMSpeshGra
         bb = bb->linear_next;
     }
 
-    /* If we found nothing we can't inline, inlining is fine. */
-    return ig;
+    return 1;
+}
 
-    /* If we can't find a way to inline, we end up here. */
-  not_inlinable:
-    MVM_free(ig->spesh_slots);
-    MVM_spesh_graph_destroy(tc, ig);
-    return NULL;
+/* Sees if it will be possible to inline the target code ref, given we could
+ * already identify a spesh candidate. Returns NULL if no inlining is possible
+ * or a graph ready to be merged if it will be possible. */
+MVMSpeshGraph * MVM_spesh_inline_try_get_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
+                                               MVMStaticFrame *target_sf,
+                                               MVMSpeshCandidate *cand,
+                                               MVMSpeshIns *invoke_ins,
+                                               char **no_inline_reason) {
+    MVMSpeshGraph *ig;
+
+    /* Check bytecode size is within the inline limit. */
+    if (cand->bytecode_size > MVM_SPESH_MAX_INLINE_SIZE &&
+            target_sf->body.bytecode_size > MVM_SPESH_MAX_INLINE_SIZE) {
+        *no_inline_reason = "bytecode is too large to inline";
+        return NULL;
+    }
+
+    /* Check the target is suitable for inlining. */
+    if (!is_static_frame_inlineable(tc, inliner, target_sf, no_inline_reason))
+        return NULL;
+
+    /* Build graph from the already-specialized bytecode and check if we can
+     * inline the graph. */
+    ig = MVM_spesh_graph_create_from_cand(tc, target_sf, cand, 0);
+    if (is_graph_inlineable(tc, inliner, target_sf, invoke_ins, ig, no_inline_reason)) {
+        /* We can inline it. Do facts discovery, which also sets usage counts.
+         * We also need to bump counts for any inline's code_ref_reg to make
+         * sure it stays available for deopt. */
+        MVMuint32 i;
+        MVM_spesh_facts_discover(tc, ig, NULL, 1);
+        for (i = 0; i < ig->num_inlines; i++) {
+            /* We can't be very precise about this, because we don't know the
+             * SSA version in effect. So bump usages of all version of the
+             * register as a conservative solution. */
+            MVMuint16 reg = ig->inlines[i].code_ref_reg;
+            MVMuint32 j;
+            for (j = 0; j < ig->fact_counts[reg]; j++)
+                MVM_spesh_usages_add_unconditional_deopt_usage(tc, ig, &(ig->facts[reg][j]));
+        }
+        return ig;
+    }
+    else {
+        /* Not possible to inline. Clear it up. */
+        MVM_free(ig->spesh_slots);
+        MVM_spesh_graph_destroy(tc, ig);
+        return NULL;
+    }
+}
+
+/* Tries to get a spesh graph for a particular unspecialized candidate. */
+MVMSpeshGraph * MVM_spesh_inline_try_get_graph_from_unspecialized(MVMThreadContext *tc,
+        MVMSpeshGraph *inliner, MVMStaticFrame *target_sf, MVMSpeshIns *invoke_ins,
+        MVMSpeshCallInfo *call_info, char **no_inline_reason) {
+    MVMSpeshGraph *ig;
+
+    /* Cannot inline with flattening args. */
+    if (call_info->cs->has_flattening) {
+        *no_inline_reason = "callsite has flattening args";
+        return NULL;
+    }
+
+    /* Check the target is suitable for inlining. */
+    if (!is_static_frame_inlineable(tc, inliner, target_sf, no_inline_reason))
+        return NULL;
+
+    /* Build the spesh graph from bytecode, transform args, do facts discovery
+     * (setting usage counts) and optimize. We do this before checking if it
+     * is inlineable as we can get rid of many :noinline ops (especially in
+     * the args specialization). */
+    ig = MVM_spesh_graph_create(tc, target_sf, 0, 1);
+    MVM_spesh_args_from_callinfo(tc, ig, call_info);
+    MVM_spesh_facts_discover(tc, ig, NULL, 0);
+    MVM_spesh_optimize(tc, ig, NULL);
+
+    /* See if it's inlineable; clean up if not. */
+    if (is_graph_inlineable(tc, inliner, target_sf, invoke_ins, ig, no_inline_reason)) {
+        return ig;
+    }
+    else {
+        MVM_free(ig->spesh_slots);
+        MVM_spesh_graph_destroy(tc, ig);
+        return NULL;
+    }
 }
 
 /* Finds the deopt index of the return. */
@@ -314,6 +378,29 @@ static void rewrite_outer_lookup(MVMThreadContext *tc, MVMSpeshGraph *g,
     new_operands[3] = code_ref_reg;
     ins->info = MVM_op_get_op(op);
     ins->operands = new_operands;
+    MVM_spesh_usages_add_by_reg(tc, g, code_ref_reg, ins);
+}
+
+static void rewrite_hlltype(MVMThreadContext *tc, MVMSpeshGraph *inlinee, MVMSpeshIns *ins) {
+    MVMObject *selected_type;
+    MVMHLLConfig *hll = inlinee->sf->body.cu->body.hll_config;
+    MVMSpeshOperand *old_ops = ins->operands;
+    MVMuint16 sslot;
+
+    switch (ins->info->opcode) {
+        case MVM_OP_hllboxtype_i: selected_type = hll->int_box_type; break;
+        case MVM_OP_hllboxtype_n: selected_type = hll->num_box_type; break;
+        case MVM_OP_hllboxtype_s: selected_type = hll->str_box_type; break;
+        case MVM_OP_hlllist:      selected_type = hll->slurpy_array_type; break;
+        case MVM_OP_hllhash:      selected_type = hll->slurpy_hash_type; break;
+        default: MVM_oops(tc, "unhandled instruction %s in rewrite_hlltype", ins->info->name);
+    }
+
+    sslot = MVM_spesh_add_spesh_slot_try_reuse(tc, inlinee, (MVMCollectable *)selected_type);
+    ins->operands = MVM_spesh_alloc(tc, inlinee, 2 * sizeof(MVMSpeshOperand));
+    ins->operands[0] = old_ops[0];
+    ins->operands[1].lit_i16 = sslot;
+    ins->info = MVM_op_get_op(MVM_OP_sp_getspeshslot);
 }
 
 /* Merges the inlinee's spesh graph into the inliner. */
@@ -327,11 +414,15 @@ MVMSpeshBB * merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
     MVMint32        i, orig_inlines, total_inlines, orig_deopt_addrs;
     MVMuint32       total_handlers = inliner->num_handlers + inlinee->num_handlers + 1;
     MVMSpeshBB     *inlinee_first_bb = NULL, *inlinee_last_bb = NULL;
+    MVMuint8        may_cause_deopt = 0;
 
     /* If the inliner and inlinee are from different compilation units, we
      * potentially have to fix up extra things. */
     MVMint32 same_comp_unit = inliner->sf->body.cu == inlinee->sf->body.cu;
 
+    MVMint32 same_hll = same_comp_unit || inliner->sf->body.cu->body.hll_config ==
+            inlinee_sf->body.cu->body.hll_config;
+
     /* Renumber the locals, lexicals, and basic blocks of the inlinee; also
      * re-write any indexes in annotations that need it. */
     MVMSpeshBB *bb = inlinee->entry;
@@ -342,7 +433,10 @@ MVMSpeshBB * merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
             MVMSpeshAnn *ann    = ins->annotations;
             while (ann) {
                 switch (ann->type) {
+                case MVM_SPESH_ANN_DEOPT_ONE_INS:
+                case MVM_SPESH_ANN_DEOPT_ALL_INS:
                 case MVM_SPESH_ANN_DEOPT_INLINE:
+                case MVM_SPESH_ANN_DEOPT_OSR:
                     ann->data.deopt_idx += inliner->num_deopt_addrs;
                     break;
                 case MVM_SPESH_ANN_INLINE_START:
@@ -353,6 +447,9 @@ MVMSpeshBB * merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
                 ann = ann->next;
             }
 
+            if (ins->info->may_cause_deopt)
+                may_cause_deopt = 1;
+
             if (opcode == MVM_SSA_PHI) {
                 for (i = 0; i < ins->info->num_operands; i++)
                     ins->operands[i].reg.orig += inliner->num_locals;
@@ -382,6 +479,11 @@ MVMSpeshBB * merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
                     if (ins->info->opcode == MVM_OP_const_s) {
                         fix_const_str(tc, inliner, inlinee, ins);
                     }
+                    if (!same_hll &&
+                            (opcode == MVM_OP_hllboxtype_i || opcode == MVM_OP_hllboxtype_n || opcode == MVM_OP_hllboxtype_s
+                             || opcode == MVM_OP_hlllist || opcode == MVM_OP_hllhash)) {
+                        rewrite_hlltype(tc, inlinee, ins);
+                    }
                 }
 
                 for (i = 0; i < ins->info->num_operands; i++) {
@@ -548,6 +650,7 @@ MVMSpeshBB * merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
     inliner->inlines[total_inlines - 1].unreachable = 0;
     inliner->inlines[total_inlines - 1].deopt_named_used_bit_field =
         inlinee->deopt_named_used_bit_field;
+    inliner->inlines[total_inlines - 1].may_cause_deopt = may_cause_deopt;
     inliner->num_inlines = total_inlines;
 
     /* Create/update per-specialization local and lexical type maps. */
@@ -648,7 +751,8 @@ MVMSpeshBB * merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
 }
 
 /* Tweak the successor of a BB, also updating the target BBs pred. */
-static void tweak_succ(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb, MVMSpeshBB *new_succ) {
+static void tweak_succ(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
+                       MVMSpeshBB *prev_pred, MVMSpeshBB *new_succ, MVMuint32 missing_ok) {
     if (bb->num_succ == 0) {
         /* It had no successors, so we'll add one. */
         bb->succ = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshBB *));
@@ -670,12 +774,12 @@ static void tweak_succ(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb, M
         MVMint32 found = 0;
         MVMint32 i;
         for (i = 0; i < new_succ->num_pred; i++)
-            if (new_succ->pred[i]->idx + 1 == new_succ->idx) {
+            if (new_succ->pred[i] == prev_pred) {
                 new_succ->pred[i] = bb;
                 found = 1;
                 break;
             }
-        if (!found)
+        if (!found && !missing_ok)
             MVM_oops(tc,
                 "Spesh inline: could not find appropriate pred to update\n");
     }
@@ -689,11 +793,14 @@ static void return_to_set(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshIns *r
     operands[1]               = return_ins->operands[0];
     return_ins->info          = MVM_op_get_op(MVM_OP_set);
     return_ins->operands      = operands;
+    MVM_spesh_get_facts(tc, g, target)->writer = return_ins;
 }
 
 static void return_to_box(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *return_bb,
                    MVMSpeshIns *return_ins, MVMSpeshOperand target,
                    MVMuint16 box_type_op, MVMuint16 box_op) {
+    MVMSpeshOperand type_temp     = MVM_spesh_manipulate_get_temp_reg(tc, g, MVM_reg_obj);
+
     /* Create and insert boxing instruction after current return instruction. */
     MVMSpeshIns      *box_ins     = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshIns));
     MVMSpeshOperand *box_operands = MVM_spesh_alloc(tc, g, 3 * sizeof(MVMSpeshOperand));
@@ -701,13 +808,19 @@ static void return_to_box(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *re
     box_ins->operands             = box_operands;
     box_operands[0]               = target;
     box_operands[1]               = return_ins->operands[0];
-    box_operands[2]               = target;
+    box_operands[2]               = type_temp;
     MVM_spesh_manipulate_insert_ins(tc, return_bb, return_ins, box_ins);
+    MVM_spesh_get_facts(tc, g, target)->writer = box_ins;
+    MVM_spesh_usages_add_by_reg(tc, g, box_operands[1], box_ins);
+    MVM_spesh_usages_add_by_reg(tc, g, box_operands[2], box_ins);
 
     /* Now turn return instruction node into lookup of appropriate box
      * type. */
+    MVM_spesh_usages_delete_by_reg(tc, g, return_ins->operands[0], return_ins);
     return_ins->info        = MVM_op_get_op(box_type_op);
-    return_ins->operands[0] = target;
+    return_ins->operands[0] = type_temp;
+    MVM_spesh_get_facts(tc, g, type_temp)->writer = return_ins;
+    MVM_spesh_manipulate_release_temp_reg(tc, g, type_temp);
 }
 
 static void rewrite_int_return(MVMThreadContext *tc, MVMSpeshGraph *g,
@@ -790,6 +903,7 @@ static void rewrite_returns(MVMThreadContext *tc, MVMSpeshGraph *inliner,
                      MVMSpeshIns *invoke_ins, MVMSpeshBB *inlinee_last_bb) {
     /* Locate return instructions. */
     MVMSpeshBB *bb = inlinee->entry;
+    MVMuint32 saw_return = 0;
     while (bb) {
         MVMSpeshIns *ins = bb->first_ins;
         while (ins) {
@@ -799,36 +913,41 @@ static void rewrite_returns(MVMThreadContext *tc, MVMSpeshGraph *inliner,
                 if (invoke_ins->info->opcode == MVM_OP_invoke_v) {
                     MVM_spesh_manipulate_insert_goto(tc, inliner, bb, ins,
                         invoke_bb->succ[0]);
-                    tweak_succ(tc, inliner, bb, invoke_bb->succ[0]);
+                    tweak_succ(tc, inliner, bb, invoke_bb, invoke_bb->succ[0], saw_return);
+                    MVM_spesh_manipulate_delete_ins(tc, inliner, bb, ins);
                 }
                 else {
                     MVM_oops(tc,
                         "Spesh inline: return_v/invoke_[!v] mismatch");
                 }
+                saw_return = 1;
                 break;
             case MVM_OP_return_i:
                 MVM_spesh_manipulate_insert_goto(tc, inliner, bb, ins,
                     invoke_bb->succ[0]);
-                tweak_succ(tc, inliner, bb, invoke_bb->succ[0]);
+                tweak_succ(tc, inliner, bb, invoke_bb, invoke_bb->succ[0], saw_return);
                 rewrite_int_return(tc, inliner, bb, ins, invoke_bb, invoke_ins);
+                saw_return = 1;
                 break;
             case MVM_OP_return_n:
                 MVM_spesh_manipulate_insert_goto(tc, inliner, bb, ins,
                     invoke_bb->succ[0]);
-                tweak_succ(tc, inliner, bb, invoke_bb->succ[0]);
+                tweak_succ(tc, inliner, bb, invoke_bb, invoke_bb->succ[0], saw_return);
                 rewrite_num_return(tc, inliner, bb, ins, invoke_bb, invoke_ins);
+                saw_return = 1;
                 break;
             case MVM_OP_return_s:
                 MVM_spesh_manipulate_insert_goto(tc, inliner, bb, ins,
                     invoke_bb->succ[0]);
-                tweak_succ(tc, inliner, bb, invoke_bb->succ[0]);
+                tweak_succ(tc, inliner, bb, invoke_bb, invoke_bb->succ[0], saw_return);
                 rewrite_str_return(tc, inliner, bb, ins, invoke_bb, invoke_ins);
                 break;
             case MVM_OP_return_o:
                 MVM_spesh_manipulate_insert_goto(tc, inliner, bb, ins,
                     invoke_bb->succ[0]);
-                tweak_succ(tc, inliner, bb, invoke_bb->succ[0]);
+                tweak_succ(tc, inliner, bb, invoke_bb, invoke_bb->succ[0], saw_return);
                 rewrite_obj_return(tc, inliner, bb, ins, invoke_bb, invoke_ins);
+                saw_return = 1;
                 break;
             }
             ins = ins->next;
@@ -867,7 +986,7 @@ static void rewrite_args(MVMThreadContext *tc, MVMSpeshGraph *inliner,
                      * argument passing instruction. */
                     ins->info = MVM_op_get_op(MVM_OP_set);
                     ins->operands[1] = arg_ins->operands[1];
-                    MVM_spesh_get_facts(tc, inliner, ins->operands[1])->usages++;
+                    MVM_spesh_usages_add_by_reg(tc, inliner, ins->operands[1], ins);
                     MVM_spesh_manipulate_delete_ins(tc, inliner,
                         call_info->prepargs_bb, arg_ins);
                     break;
@@ -875,19 +994,19 @@ static void rewrite_args(MVMThreadContext *tc, MVMSpeshGraph *inliner,
                     arg_ins->info        = MVM_op_get_op(MVM_OP_const_i64);
                     arg_ins->operands[0] = ins->operands[0];
                     MVM_spesh_manipulate_delete_ins(tc, inliner, bb, ins);
-                    MVM_spesh_get_facts(tc, inliner, arg_ins->operands[0])->usages++;
+                    MVM_spesh_usages_add_by_reg(tc, inliner, ins->operands[0], arg_ins);
                     break;
                 case MVM_OP_argconst_n:
                     arg_ins->info        = MVM_op_get_op(MVM_OP_const_n64);
                     arg_ins->operands[0] = ins->operands[0];
                     MVM_spesh_manipulate_delete_ins(tc, inliner, bb, ins);
-                    MVM_spesh_get_facts(tc, inliner, arg_ins->operands[0])->usages++;
+                    MVM_spesh_usages_add_by_reg(tc, inliner, ins->operands[0], arg_ins);
                     break;
                 case MVM_OP_argconst_s:
                     arg_ins->info        = MVM_op_get_op(MVM_OP_const_s);
                     arg_ins->operands[0] = ins->operands[0];
                     MVM_spesh_manipulate_delete_ins(tc, inliner, bb, ins);
-                    MVM_spesh_get_facts(tc, inliner, arg_ins->operands[0])->usages++;
+                    MVM_spesh_usages_add_by_reg(tc, inliner, ins->operands[0], arg_ins);
                     break;
                 default:
                     MVM_oops(tc,
@@ -939,6 +1058,19 @@ static void rewrite_args(MVMThreadContext *tc, MVMSpeshGraph *inliner,
     MVM_spesh_manipulate_delete_ins(tc, inliner, invoke_bb, call_info->prepargs_ins);
 }
 
+/* Gets the first instruction from a spesh graph, looking through multiple
+ * basic blocks to find it. */
+static MVMSpeshIns * find_first_instruction(MVMThreadContext *tc, MVMSpeshGraph *g) {
+    MVMSpeshBB *first_bb = g->entry->linear_next;
+    while (first_bb) {
+        MVMSpeshIns *first_ins = first_bb->first_ins;
+        if (first_ins)
+            return first_ins;
+        first_bb = first_bb->linear_next;
+    }
+    MVM_oops(tc, "Unexpectedly empty specialization graph");
+}
+
 /* Annotates first and last instruction in post-processed inlinee with start
  * and end inline annotations. */
 static void annotate_inline_start_end(MVMThreadContext *tc, MVMSpeshGraph *inliner,
@@ -948,18 +1080,18 @@ static void annotate_inline_start_end(MVMThreadContext *tc, MVMSpeshGraph *inlin
     /* Annotate first instruction as an inline start. */
     MVMSpeshAnn *start_ann     = MVM_spesh_alloc(tc, inliner, sizeof(MVMSpeshAnn));
     MVMSpeshAnn *end_ann       = MVM_spesh_alloc(tc, inliner, sizeof(MVMSpeshAnn));
-    MVMSpeshBB *bb             = inlinee->entry->succ[0];
-    start_ann->next            = bb->first_ins->annotations;
+    MVMSpeshIns *first_ins     = find_first_instruction(tc, inlinee);
+    start_ann->next            = first_ins->annotations;
     start_ann->type            = MVM_SPESH_ANN_INLINE_START;
     start_ann->data.inline_idx = idx;
-    bb->first_ins->annotations = start_ann;
+    first_ins->annotations = start_ann;
 
     /* Insert annotation for handler boundary indicator fixup. */
     start_ann = MVM_spesh_alloc(tc, inliner, sizeof(MVMSpeshAnn));
-    start_ann->next = bb->first_ins->annotations;
+    start_ann->next = first_ins->annotations;
     start_ann->type = MVM_SPESH_ANN_FH_START;
     start_ann->data.frame_handler_index = inline_boundary_handler;
-    bb->first_ins->annotations = start_ann;
+    first_ins->annotations = start_ann;
 
     /* Now look for last instruction and annotate it. */
     end_ann->next             = inlinee_last_bb->last_ins->annotations;
@@ -988,15 +1120,19 @@ static void annotate_inline_start_end(MVMThreadContext *tc, MVMSpeshGraph *inlin
 void MVM_spesh_inline(MVMThreadContext *tc, MVMSpeshGraph *inliner,
                       MVMSpeshCallInfo *call_info, MVMSpeshBB *invoke_bb,
                       MVMSpeshIns *invoke_ins, MVMSpeshGraph *inlinee,
-                      MVMStaticFrame *inlinee_sf, MVMSpeshOperand code_ref_reg) {
+                      MVMStaticFrame *inlinee_sf, MVMSpeshOperand code_ref_reg,
+                      MVMuint32 proxy_deopt_idx) {
+    MVMSpeshIns *first_ins;
+
     /* Merge inlinee's graph into the inliner. */
     MVMuint32 inline_boundary_handler;
     MVMSpeshBB *inlinee_last_bb = merge_graph(tc, inliner, inlinee, inlinee_sf,
         invoke_bb, invoke_ins, code_ref_reg, &inline_boundary_handler);
 
     /* If we're profiling, note it's an inline. */
-    if (inlinee->entry->linear_next->first_ins->info->opcode == MVM_OP_prof_enterspesh) {
-        MVMSpeshIns *profenter         = inlinee->entry->linear_next->first_ins;
+    first_ins = find_first_instruction(tc, inlinee);
+    if (first_ins->info->opcode == MVM_OP_prof_enterspesh) {
+        MVMSpeshIns *profenter         = first_ins;
         profenter->info                = MVM_op_get_op(MVM_OP_prof_enterinline);
         profenter->operands            = MVM_spesh_alloc(tc, inliner, sizeof(MVMSpeshOperand));
         profenter->operands[0].lit_i16 = MVM_spesh_add_spesh_slot(tc, inliner,
@@ -1017,7 +1153,17 @@ void MVM_spesh_inline(MVMThreadContext *tc, MVMSpeshGraph *inliner,
         inlinee_last_bb, inline_boundary_handler);
 
     /* Finally, turn the invoke instruction into a goto. */
+    MVM_spesh_usages_delete_by_reg(tc, inliner,
+        invoke_ins->operands[invoke_ins->info->opcode == MVM_OP_invoke_v ? 0 : 1],
+        invoke_ins);
     invoke_ins->info = MVM_op_get_op(MVM_OP_goto);
     invoke_ins->operands[0].ins_bb = inlinee->entry->linear_next;
-    tweak_succ(tc, inliner, invoke_bb, inlinee->entry->linear_next);
+    tweak_succ(tc, inliner, invoke_bb, inlinee->entry, inlinee->entry->linear_next, 0);
+
+    /* If this inline may cause deopt, then we take the deopt index at the
+     * calling point and use it as a proxy for the deopts that may happen in
+     * the inline. Otherwise, we might incorrectly fail to preserve values
+     * for the sake of deopt. */
+    if (inliner->inlines[inliner->num_inlines - 1].may_cause_deopt)
+        MVM_spesh_usages_retain_deopt_index(tc, inliner, proxy_deopt_idx);
 }
