@@ -48,7 +48,10 @@ typedef struct MVMFixedSizeAlloc MVMFixedSizeAlloc;
 typedef struct MVMFixedSizeAllocFreeListEntry MVMFixedSizeAllocFreeListEntry;
 typedef struct MVMFixedSizeAllocSafepointFreeListEntry MVMFixedSizeAllocSafepointFreeListEntry;
 typedef struct MVMFixedSizeAllocSizeClass MVMFixedSizeAllocSizeClass;
+typedef struct MVMFixedSizeAllocThread MVMFixedSizeAllocThread;
+typedef struct MVMFixedSizeAllocThreadSizeClass MVMFixedSizeAllocThreadSizeClass;
 typedef struct MVMFrame MVMFrame;
+typedef struct MVMFrameExtra MVMFrameExtra;
 typedef struct MVMFrameHandler MVMFrameHandler;
 typedef struct MVMGen2Allocator MVMGen2Allocator;
 typedef struct MVMGen2SizeClass MVMGen2SizeClass;
@@ -70,8 +73,6 @@ typedef struct MVMKnowHOWAttributeREPRBody MVMKnowHOWAttributeREPRBody;
 typedef struct MVMKnowHOWREPR MVMKnowHOWREPR;
 typedef struct MVMKnowHOWREPRBody MVMKnowHOWREPRBody;
 typedef struct MVMLexicalRegistry MVMLexicalRegistry;
-typedef struct MVMLexotic MVMLexotic;
-typedef struct MVMLexoticBody MVMLexoticBody;
 typedef struct MVMLoadedCompUnitName MVMLoadedCompUnitName;
 typedef struct MVMNFA MVMNFA;
 typedef struct MVMNFABody MVMNFABody;
@@ -79,7 +80,7 @@ typedef struct MVMNFAStateInfo MVMNFAStateInfo;
 typedef struct MVMNFGState MVMNFGState;
 typedef struct MVMNFGSynthetic MVMNFGSynthetic;
 typedef struct MVMNFGTrieNode MVMNFGTrieNode;
-typedef struct MVMNGFTrieNodeEntry MVMNGFTrieNodeEntry;
+typedef struct MVMNFGTrieNodeEntry MVMNFGTrieNodeEntry;
 typedef struct MVMNativeCall MVMNativeCall;
 typedef struct MVMNativeCallBody MVMNativeCallBody;
 typedef struct MVMNativeRef MVMNativeRef;
@@ -169,14 +170,44 @@ typedef struct MVMSpeshAnn MVMSpeshAnn;
 typedef struct MVMSpeshFacts MVMSpeshFacts;
 typedef struct MVMSpeshCode MVMSpeshCode;
 typedef struct MVMSpeshCandidate MVMSpeshCandidate;
-typedef struct MVMSpeshGuard MVMSpeshGuard;
 typedef struct MVMSpeshLogGuard MVMSpeshLogGuard;
 typedef struct MVMSpeshCallInfo MVMSpeshCallInfo;
 typedef struct MVMSpeshInline MVMSpeshInline;
+typedef struct MVMSpeshIterator MVMSpeshIterator;
+typedef struct MVMSpeshLog MVMSpeshLog;
+typedef struct MVMSpeshLogBody MVMSpeshLogBody;
+typedef struct MVMSpeshLogEntry MVMSpeshLogEntry;
+typedef struct MVMSpeshPluginState MVMSpeshPluginState;
+typedef struct MVMSpeshPluginPosition MVMSpeshPluginPosition;
+typedef struct MVMSpeshPluginGuardSet MVMSpeshPluginGuardSet;
+typedef struct MVMSpeshPluginGuard MVMSpeshPluginGuard;
+typedef struct MVMSpeshStats MVMSpeshStats;
+typedef struct MVMSpeshStatsByCallsite MVMSpeshStatsByCallsite;
+typedef struct MVMSpeshStatsByType MVMSpeshStatsByType;
+typedef struct MVMSpeshStatsType MVMSpeshStatsType;
+typedef struct MVMSpeshStatsByOffset MVMSpeshStatsByOffset;
+typedef struct MVMSpeshStatsTypeCount MVMSpeshStatsTypeCount;
+typedef struct MVMSpeshStatsInvokeCount MVMSpeshStatsInvokeCount;
+typedef struct MVMSpeshStatsTypeTupleCount MVMSpeshStatsTypeTupleCount;
+typedef struct MVMSpeshStatsPluginGuardCount MVMSpeshStatsPluginGuardCount;
+typedef struct MVMSpeshStatsStatic MVMSpeshStatsStatic;
+typedef struct MVMSpeshSimStack MVMSpeshSimStack;
+typedef struct MVMSpeshSimStackFrame MVMSpeshSimStackFrame;
+typedef struct MVMSpeshSimCallType MVMSpeshSimCallType;
+typedef struct MVMSpeshPlan MVMSpeshPlan;
+typedef struct MVMSpeshPlanned MVMSpeshPlanned;
+typedef struct MVMSpeshArgGuard MVMSpeshArgGuard;
+typedef struct MVMSpeshArgGuardNode MVMSpeshArgGuardNode;
+typedef struct MVMSpeshUsages MVMSpeshUsages;
+typedef struct MVMSpeshUseChainEntry MVMSpeshUseChainEntry;
+typedef struct MVMSpeshDeoptUseEntry MVMSpeshDeoptUseEntry;
+typedef struct MVMSpeshFrameWalker MVMSpeshFrameWalker;
 typedef struct MVMSTable MVMSTable;
 typedef struct MVMStaticFrame MVMStaticFrame;
 typedef struct MVMStaticFrameBody MVMStaticFrameBody;
 typedef struct MVMStaticFrameInstrumentation MVMStaticFrameInstrumentation;
+typedef struct MVMStaticFrameSpesh MVMStaticFrameSpesh;
+typedef struct MVMStaticFrameSpeshBody MVMStaticFrameSpeshBody;
 typedef struct MVMStorageSpec MVMStorageSpec;
 typedef struct MVMString MVMString;
 typedef struct MVMStringBody MVMStringBody;
@@ -194,7 +225,6 @@ typedef struct MVMUninstantiable MVMUninstantiable;
 typedef struct MVMWorkThread MVMWorkThread;
 typedef struct MVMIOOps MVMIOOps;
 typedef struct MVMIOClosable MVMIOClosable;
-typedef struct MVMIOEncodable MVMIOEncodable;
 typedef struct MVMIOSyncReadable MVMIOSyncReadable;
 typedef struct MVMIOSyncWritable MVMIOSyncWritable;
 typedef struct MVMIOAsyncReadable MVMIOAsyncReadable;
@@ -202,11 +232,8 @@ typedef struct MVMIOAsyncWritable MVMIOAsyncWritable;
 typedef struct MVMIOAsyncWritableTo MVMIOAsyncWritableTo;
 typedef struct MVMIOSeekable MVMIOSeekable;
 typedef struct MVMIOSockety MVMIOSockety;
-typedef struct MVMIOPipeable MVMIOPipeable;
 typedef struct MVMIOIntrospection MVMIOIntrospection;
 typedef struct MVMIOLockable MVMIOLockable;
-typedef struct MVMIOSyncStreamData MVMIOSyncStreamData;
-typedef struct MVMIOSyncPipeData MVMIOSyncPipeData;
 typedef struct MVMDecodeStream MVMDecodeStream;
 typedef struct MVMDecodeStreamBytes MVMDecodeStreamBytes;
 typedef struct MVMDecodeStreamChars MVMDecodeStreamChars;
@@ -227,13 +254,25 @@ typedef struct MVMJitGuard MVMJitGuard;
 typedef struct MVMJitInvoke MVMJitInvoke;
 typedef struct MVMJitJumpList MVMJitJumpList;
 typedef struct MVMJitControl MVMJitControl;
+typedef struct MVMJitData MVMJitData;
+typedef struct MVMJitStackSlot MVMJitStackSlot;
 typedef struct MVMJitCode MVMJitCode;
+typedef struct MVMJitCompiler MVMJitCompiler;
+typedef struct MVMJitExprTree MVMJitExprTree;
+typedef struct MVMJitExprInfo MVMJitExprInfo;
+typedef struct MVMJitExprTemplate MVMJitExprTemplate;
+typedef struct MVMJitTreeTraverser MVMJitTreeTraverser;
+typedef struct MVMJitTile MVMJitTile;
+typedef struct MVMJitTileList MVMJitTileList;
+typedef struct MVMJitTileTemplate MVMJitTileTemplate;
 typedef struct MVMProfileThreadData MVMProfileThreadData;
 typedef struct MVMProfileGC MVMProfileGC;
 typedef struct MVMProfileCallNode MVMProfileCallNode;
 typedef struct MVMProfileAllocationCount MVMProfileAllocationCount;
 typedef struct MVMProfileContinuationData MVMProfileContinuationData;
 typedef struct MVMHeapSnapshotCollection MVMHeapSnapshotCollection;
+typedef struct MVMHeapDumpIndex MVMHeapDumpIndex;
+typedef struct MVMHeapDumpIndexSnapshotEntry MVMHeapDumpIndexSnapshotEntry;
 typedef struct MVMHeapSnapshot MVMHeapSnapshot;
 typedef struct MVMHeapSnapshotType MVMHeapSnapshotType;
 typedef struct MVMHeapSnapshotStaticFrame MVMHeapSnapshotStaticFrame;
@@ -242,3 +281,10 @@ typedef struct MVMHeapSnapshotReference MVMHeapSnapshotReference;
 typedef struct MVMHeapSnapshotState MVMHeapSnapshotState;
 typedef struct MVMHeapSnapshotWorkItem MVMHeapSnapshotWorkItem;
 typedef struct MVMHeapSnapshotSeen MVMHeapSnapshotSeen;
+typedef struct MVMDebugServerBreakpointInfo MVMDebugServerBreakpointInfo;
+typedef struct MVMDebugServerBreakpointFileTable MVMDebugServerBreakpointFileTable;
+typedef struct MVMDebugServerBreakpointTable MVMDebugServerBreakpointTable;
+typedef struct MVMDebugServerHandleTable MVMDebugServerHandleTable;
+typedef struct MVMDebugServerHandleTableEntry MVMDebugServerHandleTableEntry;
+typedef struct MVMDebugServerRequestEntry MVMDebugServerRequestEntry;
+typedef struct MVMDebugServerData MVMDebugServerData;
